"""Support module to run IMAS-ParaView as a module:

.. code-block:: bash
    :caption: Options to run IMAS-ParaView CLI interface

    # Run as a module (implemented in imas_paraview/__main__.py)
    python -m imas_paraview

    # Run as "program" (see project.scripts in pyproject.toml)
    imas-paraview
"""

from imas_paraview.cli import cli

cli()



import pytest
from imas import DBEntry
from vtkmodules.vtkCommonDataModel import vtkMultiBlockDataSet

from imas_paraview.plugins.beam import BeamReader


@pytest.mark.skip(reason="no IMAS-Core available")
def test_load_beam():
    """Test if limiters are loaded in the VTK Multiblock Dataset."""
    reader = BeamReader()
    with DBEntry(
        (
            "imas:hdf5?path=/work/imas/shared/imasdb/ITER_MACHINE_DESCRIPTION/"
            "3/120000/1304/"
        ),
        "r",
    ) as entry:
        ids = entry.get("ec_launchers", lazy=True, autoconvert=False)
        reader._ids = ids
        reader.setup_ids()
        time_idx = 0
        name1 = ids.beam[0].name
        name2 = ids.beam[1].name

        # 1 selection
        output = vtkMultiBlockDataSet()
        reader._selected = [name1]
        reader._load_beam(output, time_idx)
        assert output.GetNumberOfBlocks() == 1

        # 2 selections
        output = vtkMultiBlockDataSet()
        reader._selected = [name1, name2]
        reader._load_beam(output, time_idx)
        assert output.GetNumberOfBlocks() == 2

        # All selected
        output = vtkMultiBlockDataSet()
        reader._selected = []
        for beam in ids.beam:
            reader._selected.append(beam.name)
        reader._load_beam(output, time_idx)
        assert output.GetNumberOfBlocks() == len(ids.beam)



"""Helper functions for generating paraview ServerManager XMLs for the plugins."""

import functools
from typing import List

from paraview.util.vtkAlgorithm import smdomain, smhint, smproperty


def propertygroup(label: str, properties: List[str]):
    """Convenience decorator for creating a property group.

    Args:
        label: Label to use for the property group
        properties: Names of the properties to include in the group
    """
    prop_xml = "".join(f'<Property name="{prop}" />' for prop in properties)
    return smproperty.xml(f'<PropertyGroup label="{label}">{prop_xml}</PropertyGroup>')


def enumeration(name: str, entries: dict):
    """Convenience decorator for creating an EnumerationDomain.

    Args:
        name: Name of the EnumerationDomain
        entries: Key/value pairs to use as entries in the domain
    """
    vals = "".join(f'<Entry text="{k}" value="{v}" />' for k, v in entries.items())
    return smdomain.xml(f'<EnumerationDomain name="{name}">{vals}</EnumerationDomain>')


def genericdecorator(**kwargs):
    """Convenience decorator to add a PropertyWidgetDecorator hint.

    This can be used to hide or disable property widgets based on the value of another
    property.
    """
    args = " ".join(f'{key}="{value}"' for key, value in kwargs.items())
    return smhint.xml(f'<PropertyWidgetDecorator type="GenericDecorator" {args}/>')


def stringlistdomain(property_name, **kwargs):
    """Convenience decorator to add a StringListDomain to a property.

    Args:
        property_name: Name of the StringList property with allowed values.
        kwargs: Additional keyword arguments to add as xml attributes to the domain.
    """
    args = " ".join(f'{key}="{value}"' for key, value in kwargs.items())
    return smdomain.xml(
        f"<StringListDomain {args}><RequiredProperties>"
        f'<Property name="{property_name}" function="StringInfo" />'
        "</RequiredProperties></StringListDomain>"
    )


def arrayselectionstringvector(property_name, attribute_name):
    """Convenience decorator to add a stringvector for an ArraySelectionDomain.
    Paraview requires the following functions to be defined with the corresponding
    `attribute_name`:

        - GetNumberOf{attribute_name}Arrays(self)
        - Get{attribute_name}ArrayName(self, idx)
        - Get{attribute_name}ArrayStatus(self, *args)

    Args:
        property_name: Name of the stringvector property. Must match the
            `property_name` in arrayselectiondomain
        attribute_name: Name to use in the auxiliary Paraview functions
    """
    return smproperty.xml(
        f"""
        <StringVectorProperty
        information_only="1" name="{property_name}">
            <ArraySelectionInformationHelper attribute_name="{attribute_name}" />
        </StringVectorProperty>
        """
    )


def arrayselectiondomain(property_name, **kwargs):
    """Convenience decorator to add an ArraySelectionDomain to a property."""

    def decorator(func):
        args = " ".join(f'{key}="{value}"' for key, value in kwargs.items())

        xml = f"""
        <StringVectorProperty
            {args}
            command="{func.__name__}"
            number_of_elements="0"
            number_of_elements_per_command="2"
            panel_visibility="default"
            element_types="2 0"
            repeat_command="1">
            <ArraySelectionDomain name="array_list">
                <RequiredProperties>
                    <Property function="ArrayList" name="{property_name}" />
                </RequiredProperties>
            </ArraySelectionDomain>
            <Documentation>{func.__doc__}</Documentation>
            </StringVectorProperty>
        """

        return smproperty.xml(xml)(func)

    return decorator


def checkbox(**kwargs):
    """Convenience decorator for creating a simple boolean checkbox."""

    def decorator(func):
        args = " ".join(f'{key}="{value}"' for key, value in kwargs.items())

        xml = f"""
        <IntVectorProperty
            {args}
            command="{func.__name__}"
            number_of_elements="1">
            <BooleanDomain name="bool" />
            <Documentation>{func.__doc__}</Documentation>
        </IntVectorProperty>
        """

        return smproperty.xml(xml)(func)

    return decorator


def add_docstring(func):
    """Convenience decorator to add a Documentation XML node filled with the docstring
    of the property.
    """
    # Abuse smdomain.xml, which inserts the XML element in the correct location...
    doc = func.__doc__
    if doc:
        return smdomain.xml(f"<Documentation>{doc}</Documentation>")(func)
    return func


def _propertywrapper(property, *args, **kwargs):
    """Helper function to chain smproperty.<property> and add_docstring."""

    def decorator(func):
        return property(*args, **kwargs)(add_docstring(func))

    return decorator


intvector = functools.partial(_propertywrapper, smproperty.intvector)
"""Convenience decorator that wraps `smproperty.intvector` and `add_docstring`."""
stringvector = functools.partial(_propertywrapper, smproperty.stringvector)
"""Convenience decorator that wraps `smproperty.stringvector` and `add_docstring`."""
doublevector = functools.partial(_propertywrapper, smproperty.doublevector)
"""Convenience decorator that wraps `smproperty.doublevector` and `add_docstring`."""



import numpy as np
import pytest
from imas import DBEntry
from vtk.util.numpy_support import vtk_to_numpy
from vtkmodules.vtkCommonDataModel import vtkMultiBlockDataSet

from imas_paraview.plugins.line_of_sight import LineOfSightReader
from imas_paraview.util import pol_to_cart


@pytest.mark.skip(reason="no IMAS-Core available")
def test_load_los():
    """Test if line_of_sight structures are loaded in the VTK Multiblock Dataset."""
    reader = LineOfSightReader()

    with DBEntry(
        "imas:hdf5?path=/work/imas/shared/imasdb/ITER_MACHINE_DESCRIPTION/3/150401/3/",
        "r",
    ) as entry:
        ids = entry.get("bolometer", lazy=True, autoconvert=False)
        reader._ids = ids
        reader.setup_ids()

        los1 = ids.channel[0].line_of_sight
        name1 = ids.channel[0].name
        los2 = ids.channel[1].line_of_sight
        name2 = ids.channel[1].name

        # 1 selection
        output = vtkMultiBlockDataSet()
        reader._selected = [name1]
        reader._load_los(output)
        assert output.GetNumberOfBlocks() == 1
        assert_values_match(los1, output.GetBlock(0))

        # 2 selections
        output = vtkMultiBlockDataSet()
        reader._selected = [name1, name2]
        reader._load_los(output)
        assert output.GetNumberOfBlocks() == 2
        assert_values_match(los1, output.GetBlock(0))
        assert_values_match(los2, output.GetBlock(1))

        # All selected
        output = vtkMultiBlockDataSet()
        reader._selected = []
        for channel in ids.channel:
            reader._selected.append(channel.name)
        reader._load_los(output)
        assert output.GetNumberOfBlocks() == len(ids.channel)
        for i, channel in enumerate(ids.channel):
            los = channel.line_of_sight
            assert_values_match(los, output.GetBlock(i))


@pytest.mark.skip(reason="no IMAS-Core available")
def test_load_los_ece():
    """Test if line_of_sight structures in ece are loaded in the VTK
    Multiblock Dataset."""
    reader = LineOfSightReader()
    with DBEntry(
        "imas:hdf5?path=/work/imas/shared/imasdb/ITER_MACHINE_DESCRIPTION/3/150601/22",
        "r",
    ) as entry:
        ids = entry.get("ece", lazy=True, autoconvert=False)
        reader._ids = ids
        reader.setup_ids()

        # this ece IDS does not have line_of_sight structures for each channel,
        # instead the global line_of_sight is used
        los = ids.line_of_sight
        name1 = ids.channel[0].name
        name2 = ids.channel[1].name

        # 1 selection
        output = vtkMultiBlockDataSet()
        reader._selected = [name1]
        reader._load_los(output)
        assert output.GetNumberOfBlocks() == 1
        assert_values_match(los, output.GetBlock(0))

        # 2 selections
        output = vtkMultiBlockDataSet()
        reader._selected = [name1, name2]
        reader._load_los(output)
        assert output.GetNumberOfBlocks() == 2
        assert_values_match(los, output.GetBlock(0))
        assert_values_match(los, output.GetBlock(1))

        # All selected
        output = vtkMultiBlockDataSet()
        reader._selected = []
        for channel in ids.channel:
            reader._selected.append(channel.name)
        reader._load_los(output)
        assert output.GetNumberOfBlocks() == len(ids.channel)
        for i, channel in enumerate(ids.channel):
            assert_values_match(los, output.GetBlock(i))


def assert_values_match(los, block):
    """Check if points in the line_of_sight match with the values in vtk block"""
    vtk_points = block.GetPoints()
    numpy_points = vtk_to_numpy(vtk_points.GetData())

    first_point = [
        *pol_to_cart(los.first_point.r, los.first_point.phi),
        los.first_point.z,
    ]
    second_point = [
        *pol_to_cart(los.second_point.r, los.second_point.phi),
        los.second_point.z,
    ]
    los_points = np.array([first_point, second_point])

    assert np.all(np.isclose(los_points, numpy_points))



import imas
from imas.ids_data_type import IDSDataType
from imas.ids_toplevel import IDSToplevel

from imas_paraview._ids_util import _get_nodes_from_path


def get_arrays_from_ids(
    ids,
    ggd_idx=None,
    get_empty_arrays=False,
    scalar_array_paths=None,
    vector_array_paths=None,
    create_empty_structs=False,
):
    """Fetches GGD scalar and vector arrays that reside in the IDS. If lists of
    IDSPaths of GGD arrays are provided through scalar_array_paths or
    vector_array_paths, only these will be loaded. If either of them is not provided,
    instead all GGD arrays in the IDS will be returned.

    Args:
        ids: The IDS from which to fetch GGD arrays
        ggd_idx: The GGD time step to load. Defaults to None, which corresponds with
            loading all timesteps.
        get_empty_arrays (bool): Whether to return empty GGD arrays
        scalar_array_paths: A list of IDSPaths of GGD scalar arrays to search through.
        vector_array_paths: A list of IDSPaths of GGD vector arrays to search through.
        create_empty_structs: If this flag is enabled and an empty structure is
            encountered through which has to be traversed to reach a GGD array, said
            structure is resized to have length 1.
    Returns:
        scalar_array_list: The GGD scalar arrays (real & complex)
        vector_array_list: The GGD vector arrays (normal & rphiz)
    """
    if scalar_array_paths is None or vector_array_paths is None:
        # Recursively search the IDS for GGD paths
        scalar_array_paths = []
        vector_array_paths = []
        recursive_ggd_path_search(
            ids.metadata,
            scalar_array_paths,
            vector_array_paths,
        )

    # Find scalar and vector GGD arrays in the IDS from the paths
    scalar_array_list = []
    vector_array_list = []
    for scalar_path in scalar_array_paths:
        scalar_array_list.extend(
            _get_nodes_from_path(
                ids, scalar_path, get_empty_arrays, ggd_idx, create_empty_structs
            )
        )

    for vector_path in vector_array_paths:
        vector_array_list.extend(
            _get_nodes_from_path(
                ids, vector_path, get_empty_arrays, ggd_idx, create_empty_structs
            )
        )

    return scalar_array_list, vector_array_list


def recursive_ggd_path_search(
    quantity_metadata, scalar_array_paths, vector_array_paths
):
    """Recursively searches through the metadata of an IDS node for scalar GGD arrays
    (real & complex) and vector GGD arrays (regular and rphiz), and appends the paths of
    these to the scalar_array_paths and vector_array_paths respectively.

    Args:
        quantity_metadata: The metadata of an IDS node
        scalar_array_paths: The IDSPaths of GGD scalar arrays (real & complex)
        vector_array_paths: The IDSPaths of GGD vector arrays (regular and rphiz)
    """
    for subquantity_metadata in quantity_metadata:
        if subquantity_metadata.data_type == IDSDataType.STRUCT_ARRAY:
            # Get scalar and complex scalar array quantities
            if subquantity_metadata.structure_reference in [
                "generic_grid_scalar",
                "generic_grid_scalar_complex",
            ]:
                scalar_array_paths.append(subquantity_metadata.path)

            # Get vector and rzphi-vector array quantities
            # From DDv4 onward `generic_grid_vector_components_rzphi` will be
            # replaced by `generic_grid_vector_components_rphiz`
            elif subquantity_metadata.structure_reference in [
                "generic_grid_vector_components",
                "generic_grid_vector_components_rzphi",
                "generic_grid_vector_components_rphiz",
            ]:
                vector_array_paths.append(subquantity_metadata.path)

        recursive_ggd_path_search(
            subquantity_metadata,
            scalar_array_paths,
            vector_array_paths,
        )


def create_name_recursive(node):
    """Generates a name for an IDS node. The parents of the node are
    searched recursively until the IDS toplevel is reached. The name of the metadata
    of each parent node is stored as well as the identifier, name or labels of the
    node, which are added in brackets, if applicable.

    Args:
        node: The IDS node

    Returns:
        Name of the IDS node
    """
    name_current_node = node.metadata.name
    name = ""
    if (
        "ggd" != name_current_node
        and "profiles_1d" != name_current_node
        and "profiles_2d" != name_current_node
        and "time_slice" not in name_current_node
    ):
        name_appendix = ""

        # Check if node has an identifier.name
        if hasattr(node, "identifier") and hasattr(node.identifier, "name"):
            name_appendix = str(node.identifier.name).strip()

        # Check if node has a name
        elif hasattr(node, "name"):
            name_appendix = str(node.name).strip()

        # Check if node has a label
        elif hasattr(node, "label"):
            name_appendix = str(node.label.value).strip()

        # Add identifier/name/label in between brackets to the full name
        if name_appendix != "":
            name = f"{name_current_node.capitalize()} ({name_appendix.capitalize()})"
        else:
            name = name_current_node.capitalize()

    parent = imas.util.get_parent(node)
    if parent.metadata is node.metadata:
        parent = imas.util.get_parent(parent)
    if not isinstance(parent, IDSToplevel):
        name = f"{create_name_recursive(parent)} {name}"

    name = name.strip()
    return name


def get_object_by_name(selectable, obj_name):
    """Search through a list of selectable attributes in the array selection domain
    and return the object IDS structure which matches the selected object name.

    Args:
        obj_name: Name of the object to search for.

    Returns:
        object with the corresponding name, or None if no match is found
    """
    for obj in selectable:
        if obj_name == obj.name:
            return obj
    return None



import logging
from dataclasses import dataclass

import numpy as np
import vtk
from imas.ids_struct_array import IDSStructArray
from imas.ids_structure import IDSStructure
from paraview.util.vtkAlgorithm import smhint, smproxy
from vtkmodules.util.numpy_support import numpy_to_vtk
from vtkmodules.vtkCommonDataModel import (
    vtkPartitionedDataSet,
    vtkPartitionedDataSetCollection,
)

from imas_paraview.ids_util import create_name_recursive, get_object_by_name
from imas_paraview.plugins.base_class import GGDVTKPluginBase
from imas_paraview.util import find_closest_indices

logger = logging.getLogger("imas_paraview")

# TODO: waves and distributions are currently not supported as their profiles_2d are
# stored differently
PROFILES_2D_IDS_NAMES = [
    "core_profiles",
    "equilibrium",
    "plasma_initiation",
    "plasma_profiles",
]


@dataclass
class Profile_2d:
    """Data class that stores a profiles_2d node, along with its name."""

    name: str
    profile: np.ndarray


@smproxy.source(label="2D Profiles Reader")
@smhint.xml("""<ShowInMenu category="IMAS Tools" />""")
class Profiles2DReader(GGDVTKPluginBase, is_time_dependent=True):
    """profiles_2d reader based on IMAS-Python"""

    def __init__(self):
        super().__init__("vtkPartitionedDataSetCollection", PROFILES_2D_IDS_NAMES)
        self.r = np.array([])
        self.z = np.array([])
        self._filled_profiles = []
        self._all_profiles = []

    def GetAttributeArrayName(self, idx) -> str:
        return self._selectable[idx].name

    def RequestData(self, request, inInfo, outInfo):
        if self._dbentry is None or not self._ids_and_occurrence or self._ids is None:
            return 1

        # Retrieve the selected time step and profiles
        time = self._get_selected_time_step(outInfo)
        if time is None:
            logger.warning("Selected invalid time step")
            return 1

        index_list = find_closest_indices([time], self._ids.time)
        time_idx = index_list[0]
        self.update_available_profiles(time_idx)

        if len(self._selected) > 0:
            output = vtkPartitionedDataSetCollection.GetData(outInfo)
            self._load_profiles(output)
        return 1

    def update_available_profiles(self, time_idx):
        """Searches through the profiles_2d node at the current time step for
        available profiles to select. Which profiles are shown in the array domain
        selector is based on whether the "Show All" checkbox is enabled.

        Args:
            time_idx: The time index of the 2D profile
        """
        if self._ids is None:
            logger.error("Could not find the IDS.")
            return

        if self._ids.metadata.name == "equilibrium":
            if not time_idx < len(self._ids.time_slice):
                logger.error(f"There is no profiles_2d at time step {time_idx}")
                return
            for profile in self._ids.time_slice[time_idx].profiles_2d:
                self._search_valid_profile(profile)
                # Exit once a valid profiles_2d node has been found
                if self._filled_profiles:
                    break
        else:
            if not time_idx < len(self._ids.profiles_2d):
                logger.error(f"There is no profiles_2d at time step {time_idx}")
                return
            profile = self._ids.profiles_2d[time_idx]
            self._search_valid_profile(profile)

        if not self._filled_profiles:
            logger.error("Could not find a valid profiles_2d node.")
            return

        if self.show_all:
            self._selectable = self._get_selectable_profiles(self._all_profiles)
        else:
            self._selectable = self._get_selectable_profiles(self._filled_profiles)

    def _search_valid_profile(self, profile):
        """Looks for valid profiles within a profiles_2d node and stores them into a
        list.

        Args:
            profile: The profile node at a specific time step.
        """
        self.r = np.array([])
        self.z = np.array([])
        self._filled_profiles = []
        self._all_profiles = []

        if (
            hasattr(profile, "r")
            and hasattr(profile, "z")
            and profile.r.has_value
            and profile.r.shape == profile.z.shape
        ):
            # Use the 2D R and Z arrays to construct the grid
            self.r = profile.r
            self.z = profile.z
        else:
            # Check if this is a rectangular grid, and use dim1/dim2
            grid_type = profile.grid_type.index
            if grid_type != 1:
                logger.debug(
                    f"Found a grid type with identifier index of {grid_type}. "
                    "Only rectangular profiles (index = 1) are supported."
                )
                return
            self.r = np.tile(profile.grid.dim1, (len(profile.grid.dim2), 1))
            self.z = np.tile(profile.grid.dim2, (len(profile.grid.dim1), 1)).T

        # If the grid is not filled, continue to the next profiles_2d
        if len(self.r) == 0 or len(self.z) == 0:
            return

        self._recursively_find_profiles(profile)

    def setup_ids(self):
        """
        Select which profiles to show in the array domain selector, based
        on whether the "Show All" checkbox is enabled.
        """
        # WARN: The selected time cannot be fetched during the RequestInformation
        # step, so we take the first time index here to update the domain selection
        # array. This causes some issues if there are profiles in the IDS which are
        # not filled in later time steps. For example, if the ion for
        # profiles_1d[0]/ion[0].density has name "A", but for
        # profiles_1d[1]/ion[0].density the ion has the name "B", this profile is
        # lost. To mitigate this, ensure that the ions which will be used are
        # all defined at the first time step. Alternatively, a current workaround
        # for this is to press the "Show All" checkbox twice (i.e. enable and then
        # disable), when a later time step is selected. This will cause the UI
        # to update and show the available profiles at the selected time step.
        time_idx = 0
        self.update_available_profiles(time_idx)

    def _get_selectable_profiles(self, profiles):
        """Creates a list of Profile_2d objects with generated names.

        Args:
            profiles: A list of profile objects to be processed.

        Returns:
            A list of `Profile_2d` objects created from the provided profiles
        """
        profile_list = []
        for profile in profiles:
            name = create_name_recursive(profile)
            profile = Profile_2d(name, profile)
            profile_list.append(profile)
        return profile_list

    def _recursively_find_profiles(self, node):
        """Recursively traverses through the IDS node searching for filled 2d profiles.
        If a filled profile is found, it is appended to self._filled_profiles, all
        profiles are appended to self._all_profiles, filled or not.

        Args:
            node: the node to search through.
        """

        if node.metadata.name in ("grid", "r", "z"):
            return
        elif isinstance(node, IDSStructure) or isinstance(node, IDSStructArray):
            for subnode in node:
                self._recursively_find_profiles(subnode)
        else:
            if node.metadata.ndim == 2:
                if node.has_value:
                    self._filled_profiles.append(node)
                self._all_profiles.append(node)

    def _load_profiles(self, output):
        """Go through the list of selected profiles, and load each of them in a
        separate vtkUnstructuredGrid object, which are all combined into a
        vtkPartitionedDataSetCollection.

        Args:
            output: The vtkPartitionedDataSetCollection containing the profiles.
        """

        # Since profiles share the same grid, it can be converted once
        vtk_points = self._create_vtkpoints()

        for i, profile_name in enumerate(self._selected):
            profile = get_object_by_name(self._selectable, profile_name)
            if profile is None:
                raise ValueError(f"Could not find {profile_name}")

            logger.info(f"Selected {profile.name}")

            vtk_scalars = self._create_vtkscalars(profile)
            vtk_ugrid = self._create_ugrid(vtk_points, vtk_scalars)

            partitioned_dataset = vtkPartitionedDataSet()
            partitioned_dataset.SetPartition(0, vtk_ugrid)

            output.SetPartitionedDataSet(i, partitioned_dataset)

    def _create_vtkpoints(self):
        """Create vtkPoints containing the radial and height coordinates of the profile.

        Returns:
            The VTK points.
        """
        r_flat = self.r.ravel()
        z_flat = self.z.ravel()
        points = np.column_stack((r_flat, np.zeros_like(r_flat), z_flat))

        vtk_points = vtk.vtkPoints()
        vtk_points.SetData(numpy_to_vtk(points.astype(np.float64), deep=True))
        return vtk_points

    def _create_vtkscalars(self, profile):
        """Create VTK array containing the values of the profile at the grid.

        Args:
            profile: The profile to load the values from

        Returns:
            The converted VTK array.
        """
        values_flat = profile.profile.ravel()
        vtk_scalars = numpy_to_vtk(values_flat.astype(np.float64), deep=True)
        units = profile.profile.metadata.units
        vtk_scalars.SetName(f"{profile.name} [{units}]")
        return vtk_scalars

    def _create_ugrid(self, vtk_points, vtk_scalars):
        """Create a vtkUnstructuredGrid of the given profile.

        Args:
            profile: The profile to create a ugrid for.

        Returns:
            The created unstructured grid.
        """

        ugrid = vtk.vtkUnstructuredGrid()
        ugrid.GetPointData().SetScalars(vtk_scalars)

        num_points = vtk_points.GetNumberOfPoints()
        cells = vtk.vtkCellArray()

        for i in range(num_points):
            vertex = vtk.vtkVertex()
            vertex.GetPointIds().SetId(0, i)
            cells.InsertNextCell(vertex)

        ugrid = vtk.vtkUnstructuredGrid()
        ugrid.SetPoints(vtk_points)
        ugrid.GetPointData().SetScalars(vtk_scalars)
        ugrid.SetCells(vtk.VTK_VERTEX, cells)
        return ugrid



"""Plugin to view profiles_1D nodes"""

import logging
from dataclasses import dataclass

import numpy as np
from imas.ids_struct_array import IDSStructArray
from imas.ids_structure import IDSStructure
from paraview.util.vtkAlgorithm import smhint, smproxy
from vtkmodules.util.numpy_support import numpy_to_vtk
from vtkmodules.vtkCommonDataModel import vtkTable

from imas_paraview.ids_util import create_name_recursive, get_object_by_name
from imas_paraview.plugins.base_class import GGDVTKPluginBase
from imas_paraview.util import find_closest_indices

logger = logging.getLogger("imas_paraview")

PROFILES_1D_IDS_NAMES = ["core_profiles", "core_sources"]


@dataclass
class Profile_1d:
    """Data class that stores 1d profiles, along with its name and coordinate array."""

    name: str
    coordinates: np.ndarray
    profile: np.ndarray


@smproxy.source(label="1D Profiles Reader")
@smhint.xml("""<ShowInMenu category="IMAS Tools" />""")
class Profiles1DReader(GGDVTKPluginBase, is_time_dependent=True):
    """profiles_1d reader based on IMAS-Python"""

    def __init__(self):
        super().__init__("vtkTable", PROFILES_1D_IDS_NAMES)

    def GetAttributeArrayName(self, idx) -> str:
        return self._selectable[idx].name

    def RequestData(self, request, inInfo, outInfo):
        if self._dbentry is None or not self._ids_and_occurrence or self._ids is None:
            return 1

        # Retrieve the selected time step and profiles
        time = self._get_selected_time_step(outInfo)
        if time is None:
            logger.warning("Selected invalid time step")
            return 1

        index_list = find_closest_indices([time], self._ids.time)
        time_idx = index_list[0]
        self.update_available_profiles(time_idx)

        if len(self._selected) > 0:
            output = vtkTable.GetData(outInfo)
            self._load_profiles(output)
        return 1

    def _load_profiles(self, output):
        """Creates vtkDoubleArrays for the selected profiles and stores them into
        a vtkTable. If multiple profiles are selected, it is checked if their
        coordinates match, if they do they can be plotted in the same 1d plot.

        Args:
            output: vtkTable output of the plugin containing the profiles and their
                coordinates as columns.
        """
        prev_x = None
        for profile_name in self._selected:
            profile = get_object_by_name(self._selectable, profile_name)
            if profile is None:
                logger.warning(
                    f"Could not find a matching profile with name {profile_name}"
                )
                continue

            if len(profile.profile) == 0:
                logger.warning(f"The selected profile {profile.name} is empty.")
                continue
            if len(profile.coordinates) != len(profile.profile):
                logger.warning(
                    "The length of the linked coordinate array does not match."
                )
                continue

            logger.info(f"Selected {profile_name}.")
            y_values = self._create_vtk_double_array(profile.profile, profile.name)
            output.AddColumn(y_values)

            # If multiple profiles are selected, check if their coordinates match
            if prev_x is None:
                x_values = self._create_vtk_double_array(
                    profile.coordinates, profile.coordinates.metadata.name
                )
                output.AddColumn(x_values)
            else:
                if profile.coordinates is not prev_x:
                    raise RuntimeError(
                        "The X values for the selected profiles do not match. "
                        "Select the profiles one by one instead."
                    )
            prev_x = profile.coordinates

    def _create_vtk_double_array(self, values, name):
        """Creates a vtkDoubleArray with the given name and values.

        Args:
            values: values to store in the array
            name: name to give to the array

        Returns:
            vtkDoubleArray with given name and values
        """
        vtk_array = numpy_to_vtk(values, deep=1)
        vtk_array.SetName(name)
        return vtk_array

    def update_available_profiles(self, time_idx):
        """
        Searches through the profiles_1d node at the current time step for
        available profiles to select. Which profiles show in the array domain selector
        is based on whether the "Show All" checkbox is enabled.
        """

        if self._ids is not None:
            self._filled_profiles = []
            self._all_profiles = []
            if self._ids.metadata.name == "core_profiles":
                for profile_node in self._ids.profiles_1d[time_idx]:
                    self._recursive_find_profiles(profile_node)
            elif self._ids.metadata.name == "core_sources":
                for source in self._ids.source:
                    for profile_node in source.profiles_1d[time_idx]:
                        self._recursive_find_profiles(profile_node)
            else:
                raise NotImplementedError(
                    "Currently only the 1D profiles of the 'core_profiles' and "
                    "'core_sources' are supported"
                )
        if self.show_all:
            self._selectable = self._get_profiles(self._all_profiles)
        else:
            self._selectable = self._get_profiles(self._filled_profiles)

    def setup_ids(self):
        """
        Select which profiles to show in the array domain selector, based
        on whether the "Show All" checkbox is enabled.
        """
        # WARN: The selected time cannot be fetched during the RequestInformation
        # step, so we take the first time index here to update the domain selection
        # array. This causes some issues if there are profiles in the IDS which are
        # not filled in later time steps. For example, if the ion for
        # profiles_1d[0]/ion[0].density has name "A", but for
        # profiles_1d[1]/ion[0].density the ion has the name "B", this profile is
        # lost. To mitigate this, ensure that the ions which will be used are
        # all defined at the first time step. Alternatively, a current workaround
        # for this is to press the "Show All" checkbox twice (i.e. enable and then
        # disable), when a later time step is selected. This will cause the UI
        # to update and show the available profiles at the selected time step.
        time_idx = 0
        self.update_available_profiles(time_idx)

    def _get_profiles(self, profiles):
        """Filters and processes a list of profiles to extract those containing
        coordinates and creates a list of Profile_1d objects with generated names.

        Args:
            profiles: A list of profile objects to be processed.

        Returns:
            A list of `Profile_1d` objects created from the provided profiles that
                  contain valid coordinates.
        """
        profile_list = []
        for profile in profiles:
            # Only store the profile if it contains coordinates
            if profile.metadata.coordinate1.references:
                path = profile.metadata.coordinate1.references[0]
                coordinates = path.goto(profile)
                name = create_name_recursive(profile)

                profile = Profile_1d(name, coordinates, profile)
                profile_list.append(profile)
        return profile_list

    def _recursive_find_profiles(self, node):
        """Recursively traverses through the IDS node searching for filled 1d profiles.
        If a filled profile is found, it is appended to self._filled_profiles.

        Args:
            node: the node to search through.
        """
        if isinstance(node, IDSStructure) or isinstance(node, IDSStructArray):
            for subnode in node:
                self._recursive_find_profiles(subnode)
        else:
            if node.metadata.ndim == 1:
                if node.has_value:
                    self._filled_profiles.append(node)
                self._all_profiles.append(node)



"""imas plugin to view positional data of barometry gauges and embedded langmuir
probes."""

import logging
from dataclasses import dataclass

from imas.ids_structure import IDSStructArray, IDSStructure
from paraview.util.vtkAlgorithm import smhint, smproxy
from vtkmodules.vtkCommonCore import vtkPoints
from vtkmodules.vtkCommonDataModel import vtkPolyData

from imas_paraview.ids_util import get_object_by_name
from imas_paraview.plugins.base_class import GGDVTKPluginBase
from imas_paraview.util import pol_to_cart

logger = logging.getLogger("imas_paraview")

SUPPORTED_IDS_NAMES = ["barometry", "langmuir_probes", "magnetics"]


@dataclass
class PositionStructure:
    """Data class that stores positional IDS structures, along with its name. For
    barometry the structures are gauges, for langmuire probes, the structures are
    embedded probes"""

    name: str
    position_structure: IDSStructure


@smproxy.source(label="Position Reader")
@smhint.xml("""<ShowInMenu category="IMAS Tools" />""")
class PositionReader(GGDVTKPluginBase):
    """Positions reader based on IMAS-Python"""

    def __init__(self):
        super().__init__("vtkPolyData", SUPPORTED_IDS_NAMES)

    def GetAttributeArrayName(self, idx) -> str:
        return self._selectable[idx].name

    def RequestData(self, request, inInfo, outInfo):
        if self._dbentry is None or not self._ids_and_occurrence or self._ids is None:
            return 1

        if len(self._selected) > 0:
            output = vtkPolyData.GetData(outInfo)
            self._load_position(output)
        return 1

    def setup_ids(self):
        """
        Select which probes or gauges to show in the array domain selector. The names
        from the gauges or probes are to the array domain selector.
        """
        assert self._ids is not None, "IDS cannot be empty during setup."

        self._selectable = []

        if self._ids.metadata.name == "barometry":
            aos_list = [self._ids.gauge]
        elif self._ids.metadata.name == "langmuir_probes":
            aos_list = [self._ids.embedded]
        elif self._ids.metadata.name == "magnetics":
            aos_list = []
            # Depending on the DD version, some attributes might not exist
            attributes = [
                "flux_loop",
                "bpol_probe",
                "b_field_pol_probe",
                "b_field_tor_probe",
                "rogowski_coil",
                "b_field_phi_probe",
            ]
            for attr in attributes:
                if hasattr(self._ids, attr):
                    aos_list.append(getattr(self._ids, attr))

        else:
            raise NotImplementedError(
                f"Unable to find load position for {self._ids.metadata.name}."
            )
        for aos in aos_list:
            for i, structure in enumerate(aos):
                name = structure.name
                if name == "":
                    name = f"device {i}"
                    logger.warning(
                        f"Found a device without a name, it will be loaded as {name}."
                    )

                if hasattr(structure, "identifier"):
                    identifier = structure.identifier
                    if not identifier == "":
                        name = f"{name} / {identifier}"

                selectable = PositionStructure(name, structure)
                self._selectable.append(selectable)

    def _load_position(self, output):
        """Go through the list of selected position structures, and load each of them
        in a vtkPoints object, which are all saved into a vtkPolyData object.

        Args:
            output: The vtkPolyData containing the positions.
        """
        points = vtkPoints()
        for name in self._selected:
            pos_struct = get_object_by_name(self._selectable, name)

            if pos_struct is None:
                raise ValueError(f"Could not find {name}")

            pos = pos_struct.position_structure.position
            logger.info(f"Selected {pos_struct.name}")

            if isinstance(pos, IDSStructArray):
                for pos_struct in pos:
                    pos_cart = (
                        *pol_to_cart(pos_struct.r, pos_struct.phi),
                        pos_struct.z,
                    )
                    points.InsertNextPoint(*pos_cart)
            else:
                pos_cart = (*pol_to_cart(pos.r, pos.phi), pos.z)
                points.InsertNextPoint(*pos_cart)

        output.SetPoints(points)



import logging

import imas
import numpy as np

from imas_paraview.ids_util import get_arrays_from_ids
from imas_paraview.util import create_first_ggd, create_first_grid, int32array

logger = logging.getLogger("imas_paraview")


def fill_NxN_grid(grid_ggd, N):
    """Fills the grid_ggd of an IDS with a uniform rectangular grid of size N x N,
    containing vertices, edges and faces.

    Adapted from https://sharepoint.iter.org/departments/POP/CM/IMDesign/Data%20Model/sphinx/3.41/ggd_guide/examples.html # noqa

    Args:
        grid_ggd: The GGD grid that will be filled with the N x N grid.
        N: The size of the N x N grid

    Returns:
        num_vertices: The number of vertices in the generated grid_ggd
        num_edges: The number of edges in the generated grid_ggd
        num_faces: The number of faces in the generated grid_ggd
    """

    # Set grid
    grid_ggd.identifier.name = "linear"
    grid_ggd.identifier.index = 1
    grid_ggd.identifier.description = f"A simple {N} x {N} rectangular grid"

    # Set space
    grid_ggd.space.resize(1)
    space = grid_ggd.space[0]
    space.identifier.name = "primary_standard"
    space.identifier.index = 1
    space.identifier.description = "Primary space defining the standard grid"
    space.geometry_type.index = 0
    space.coordinates_type.resize(2)
    space.coordinates_type[0] = 1
    space.coordinates_type[1] = 2

    space.objects_per_dimension.resize(3)
    num_vertices, num_edges, num_faces = build_grid(N, space)

    # Create subset
    grid_ggd.grid_subset.resize(3)
    grid_subsets = grid_ggd.grid_subset
    grid_subsets[0].dimension = 1
    grid_subsets[0].identifier.name = "vertices"
    grid_subsets[0].identifier.index = 1
    grid_subsets[0].identifier.description = "All vertices in the domain"

    grid_subsets[1].dimension = 2
    grid_subsets[1].identifier.name = "edges"
    grid_subsets[1].identifier.index = 2
    grid_subsets[1].identifier.description = "All edges in the domain"

    grid_subsets[2].dimension = 3
    grid_subsets[2].identifier.name = "faces"
    grid_subsets[2].identifier.index = 5
    grid_subsets[2].identifier.description = "All faces in the domain"

    # Create elements for vertices
    grid_subsets[0].element.resize(num_vertices)
    for i, element in enumerate(grid_subsets[0].element):
        element.object.resize(1)
        element.object[0].space = 1
        element.object[0].dimension = 1
        element.object[0].index = i + 1

    # Create elements for edges
    grid_subsets[1].element.resize(num_edges)
    for i, element in enumerate(grid_subsets[1].element):
        element.object.resize(1)
        element.object[0].space = 1
        element.object[0].dimension = 2
        element.object[0].index = i + 1

    # Create elements for faces
    grid_subsets[2].element.resize(num_faces)
    for i, element in enumerate(grid_subsets[2].element):
        element.object.resize(1)
        element.object[0].space = 1
        element.object[0].dimension = 3
        element.object[0].index = i + 1

    return num_vertices, num_edges, num_faces


def build_grid(N, space):
    """Builds the vertices, edges and faces for an N x N grid.

    Args:
        N: Size of the grid
        space: Space AoS of the GGD grid

    Returns:
        num_vertices: The total number of vertices
        num_edges: The total number of edges
        num_faces: The total number of faces
    """
    num_vertices = set_vertices(N, space)
    num_edges = set_edges(N, space)
    num_faces = set_faces(N, space)
    return num_vertices, num_edges, num_faces


def set_vertices(N, space):
    """Sets the vertices for an N x N grid.

    Args:
        N: Size of the grid
        space: Space AoS of the GGD grid

    Returns:
        num_vertices: The total number of vertices
    """
    vertices = space.objects_per_dimension[0].object

    num_vertices = N * N
    vertices.resize(num_vertices)

    for i in range(N):
        for j in range(N):
            idx = i * N + j
            vertices[idx].geometry = [float(j), float(i)]
    return num_vertices


def set_edges(N, space):
    """Sets the edges for an N x N grid.

    Args:
        N: Size of the grid
        space: Space AoS of the GGD grid

    Returns:
        num_edges: The total number of edges
    """
    edges = space.objects_per_dimension[1].object

    num_edges = 2 * (N - 1) * N  # Total number of edges (horizontal + vertical)
    edges.resize(num_edges)

    edge_idx = 0

    # Set horizontal edges (left to right connections)
    for i in range(N):
        for j in range(N - 1):
            edges[edge_idx].nodes = int32array([i * N + j + 1, i * N + j + 2])
            edge_idx += 1

    # Set vertical edges (top to bottom connections)
    for i in range(N - 1):
        for j in range(N):
            edges[edge_idx].nodes = int32array([i * N + j + 1, (i + 1) * N + j + 1])
            edge_idx += 1
    return num_edges


def set_faces(N, space):
    """Sets the faces for an N x N grid.

    Args:
        N: Size of the grid
        space: Space AoS of the GGD grid

    Returns:
        num_faces: The total number of faces
    """
    face = space.objects_per_dimension[2].object

    num_faces = (N - 1) * (N - 1)  # Total number of faces (squares)
    face.resize(num_faces)

    face_idx = 0

    # Set faces for each square in the grid
    for i in range(N - 1):
        for j in range(N - 1):
            top_left = i * N + j + 1
            top_right = top_left + 1
            bottom_left = top_left + N
            bottom_right = bottom_left + 1
            face[face_idx].nodes = int32array(
                [top_left, top_right, bottom_right, bottom_left]
            )
            face_idx += 1
    return num_faces


def fill_vector_quantity(vector_quantity, num_vertices, num_edges, num_faces):
    """Fills vector quantity with with random data for each vertex, edge and face.
    Only the radial, poloidal and toroidal components of the vector quantity are filled.

    Args:
        vector_quantity: The vector quantity to be filled
        num_vertices: The number of vertices in the grid_ggd
        num_edges: The number of edges in the grid_ggd
        num_faces: The number of faces in the grid_ggd
    """
    # Allocate memory for 3 entries: vertices, edges and faces
    vector_quantity.resize(3)

    # Fill values for vertices
    vector_quantity[0].grid_index = 1
    vector_quantity[0].grid_subset_index = 1

    vector_quantity[0].radial = np.random.rand(num_vertices)
    vector_quantity[0].z = np.random.rand(num_vertices)

    # Fill values for edges
    vector_quantity[1].grid_index = 1
    vector_quantity[1].grid_subset_index = 2
    vector_quantity[1].radial = np.random.rand(num_edges)
    vector_quantity[1].z = np.random.rand(num_edges)

    # Fill values for faces
    vector_quantity[2].grid_index = 1
    vector_quantity[2].grid_subset_index = 5
    vector_quantity[2].radial = np.random.rand(num_faces)
    vector_quantity[2].z = np.random.rand(num_faces)


def fill_vector_rzphi_quantity(vector_quantity, num_vertices, num_edges, num_faces):
    """Fills vector rzphi quantity with with random data for each vertex, edge and face.
    Only the radial, toroidal and z components of the vector quantity are filled.

    Args:
        vector_quantity: The vector quantity to be filled
        num_vertices: The number of vertices in the grid_ggd
        num_edges: The number of edges in the grid_ggd
        num_faces: The number of faces in the grid_ggd
    """
    # Allocate memory for 3 entries: vertices, edges and faces
    vector_quantity.resize(3)

    # Fill values for vertices
    vector_quantity[0].grid_index = 1
    vector_quantity[0].grid_subset_index = 1
    vector_quantity[0].r = np.random.rand(num_vertices)
    vector_quantity[0].z = np.random.rand(num_vertices)

    # Fill values for edges
    vector_quantity[1].grid_index = 1
    vector_quantity[1].grid_subset_index = 2
    vector_quantity[1].r = np.random.rand(num_edges)
    vector_quantity[1].z = np.random.rand(num_edges)

    # Fill values for faces
    vector_quantity[2].grid_index = 1
    vector_quantity[2].grid_subset_index = 5
    vector_quantity[2].r = np.random.rand(num_faces)
    vector_quantity[2].z = np.random.rand(num_faces)


def fill_scalar_quantity(scalar_quantity, num_vertices, num_edges, num_faces):
    """Fills scalar quantity with random data for each vertex, edge and face.

    Args:
        scalar_quantity: The scalar quantity to be filled
        num_vertices: The number of vertices in the grid_ggd
        num_edges: The number of edges in the grid_ggd
        num_faces: The number of faces in the grid_ggd
    """
    # Allocate memory for 3 entries: vertices, edges and faces
    scalar_quantity.resize(3)

    # Set 6 vertices
    scalar_quantity[0].grid_index = 1
    scalar_quantity[0].grid_subset_index = 1
    scalar_quantity[0].values = np.random.rand(num_vertices)

    # Set 7 edges
    scalar_quantity[1].grid_index = 1
    scalar_quantity[1].grid_subset_index = 2
    scalar_quantity[1].values = np.random.rand(num_edges)

    # Set 2 faces
    scalar_quantity[2].grid_index = 1
    scalar_quantity[2].grid_subset_index = 5
    scalar_quantity[2].values = np.random.rand(num_faces)


def _generate_random_complex(num_entries):
    """Generates a list of random complex numbers.

    Args:
        num_entries: The length of the returned list

    Returns:
        a list of random complex numbers
    """
    return [np.random.rand() + 1j * np.random.rand() for _ in range(num_entries)]


def fill_complex_scalar_quantity(
    complex_scalar_quantity, num_vertices, num_edges, num_faces
):
    """Fills complex scalar quantity with random data for each vertex, edge and face.

    Args:
        complex_scalar_quantity: The scalar quantity to be filled
        num_vertices: The number of vertices in the grid_ggd
        num_edges: The number of edges in the grid_ggd
        num_faces: The number of faces in the grid_ggd
    """
    # Allocate memory for 3 entries: vertices, edges and faces
    complex_scalar_quantity.resize(3)

    # Set vertices
    complex_scalar_quantity[0].grid_index = 1
    complex_scalar_quantity[0].grid_subset_index = 1
    complex_scalar_quantity[0].values = _generate_random_complex(num_vertices)

    # Set edges
    complex_scalar_quantity[1].grid_index = 1
    complex_scalar_quantity[1].grid_subset_index = 2
    complex_scalar_quantity[1].values = _generate_random_complex(num_edges)

    # Set faces
    complex_scalar_quantity[2].grid_index = 1
    complex_scalar_quantity[2].grid_subset_index = 5
    complex_scalar_quantity[2].values = _generate_random_complex(num_faces)


def fill_ggd_data(ids, num_vertices, num_edges, num_faces):
    """Fills all generic grid scalar and generic grid vector components of a GGD
    with random values.

    Args:
        ids: The IDS for which the GGD will be filled
        num_vertices: The number of vertices in the grid_ggd
        num_edges: The number of edges in the grid_ggd
        num_faces: The number of faces in the grid_ggd
    """

    # Fill IDS structure with random values
    scalar_array_list, vector_array_list = get_arrays_from_ids(
        ids, get_empty_arrays=True, create_empty_structs=True
    )

    # Read scalar arrays
    for scalar_array in scalar_array_list:
        if scalar_array.metadata.structure_reference == "generic_grid_scalar":
            fill_scalar_quantity(scalar_array, num_vertices, num_edges, num_faces)
        elif scalar_array.metadata.structure_reference == "generic_grid_scalar_complex":
            fill_complex_scalar_quantity(
                scalar_array, num_vertices, num_edges, num_faces
            )

    # Read vector arrays
    for vector_array in vector_array_list:
        structure_reference = vector_array.metadata.structure_reference
        if structure_reference == "generic_grid_vector_components":
            fill_vector_quantity(vector_array, num_vertices, num_edges, num_faces)
        # From DDv4 onward `generic_grid_vector_components_rzphi` will be
        # replaced by `generic_grid_vector_components_rphiz`
        elif structure_reference in [
            "generic_grid_vector_components_rzphi",
            "generic_grid_vector_components_rphiz",
        ]:
            fill_vector_rzphi_quantity(vector_array, num_vertices, num_edges, num_faces)


def fill_ids(ids, time_steps=1, grid_size=2):
    """Fills the IDS with an N x N uniform GGD grid and fills all GGD arrays on this
    grid with random values.

    Args:
        ids: IDS to be filled.
        time_steps: Number of time steps to create in the IDS.
        grid_size: Size of the N x N grid. Defaults to 2, meaning a 2 x 2 grid.
    """

    # Create an empty grid_ggd
    grid_ggd = create_first_grid(ids)

    # Skip filling grid_ggd if it does not exist
    if grid_ggd is None:
        logger.warning(f"{ids.metadata.name} has no grid_ggd")
    else:
        # Create time steps
        ids.time = [float(t) for t in range(time_steps)]
        ids.ids_properties.homogeneous_time = imas.ids_defs.IDS_TIME_MODE_HOMOGENEOUS

        # Create grid and GGD AoS
        grid_ggd_aos = imas.util.get_parent(grid_ggd)

        grid_ggd_aos.resize(time_steps)
        ggd = create_first_ggd(ids)
        ggd_aos = imas.util.get_parent(ggd)
        ggd_aos.resize(time_steps)

        # Create uniform grids and fill them with random GGD data
        for i in range(time_steps):
            num_vertices, num_edges, num_faces = fill_NxN_grid(
                grid_ggd_aos[i], grid_size
            )
            logger.debug(f"filled grid_ggd at index {i}.")
        fill_ggd_data(ids, num_vertices, num_edges, num_faces)

    fill_ids_specific(ids)


def fill_ids_specific(ids):
    """Fill IDS-specific nodes such that these pass the IDS validation check.

    Args:
        ids: IDS to be filled.
    """
    ids_name = ids.metadata.name
    if ids_name == "wall":
        ids.description_ggd[0].thickness.resize(len(ids.time))
    elif ids_name == "runaway_electrons":
        ids.ggd_fluid.resize(len(ids.time))



import logging
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path

from vtk import vtkXMLPartitionedDataSetCollectionWriter
from vtkmodules.vtkCommonCore import vtkPoints
from vtkmodules.vtkCommonDataModel import (
    vtkCompositeDataSet,
    vtkDataAssembly,
    vtkPartitionedDataSetCollection,
)

from imas_paraview.io import read_geom, read_jorek, read_ps
from imas_paraview.util import find_closest_indices, get_grid_ggd

logger = logging.getLogger("imas_paraview")


# Make dataclass frozen so that it is hashable, which is required for grid caching
@dataclass(frozen=True)
class InterpSettings:
    """Data class containing Fourier interpolation settings."""

    n_plane: int = 0
    phi_start: float = 0.0
    phi_end: float = 0.0


class Converter:
    def __init__(self, ids, dbentry=None):
        self.ids = ids
        self.dbentry = dbentry
        self.time_idx = None
        self.grid_ggd = None
        self.output = None
        self.ps_reader = None
        self.reference_grid_path = None
        self.get_grids = lru_cache(maxsize=32)(self.get_grids)

    def write_to_xml(self, output_path: Path, index_list=[0]):
        """Convert an IDS to VTK format and write it to disk using the XML output
        writer.

        Args:
            output_path: Path for the output directory.
            index_list: A list of time indices to convert. By default only the first
                time step is converted.
        """
        logger.info(f"Creating a output directory at {output_path}")
        output_path.mkdir(parents=True, exist_ok=True)
        output_path = output_path / self.ids.metadata.name
        if any(index >= len(self.ids.time) for index in index_list):
            raise RuntimeError("A provided index is out of bounds.")

        for index in index_list:
            logger.info(f"Converting time step {self.ids.time[index]}...")
            vtk_object = self.ggd_to_vtk(time_idx=index)
            if vtk_object is None:
                logger.warning(f"Could not convert GGD at time index {index} to VTK.")
                continue
            self._write_vtk_to_xml(vtk_object, Path(f"{output_path}_{index}"))

    def ggd_to_vtk(
        self,
        time=None,
        time_idx=None,
        scalar_paths=None,
        vector_paths=None,
        plane_config: InterpSettings = InterpSettings(),
        outInfo=None,
        progress=None,
        parent_idx=0,
    ):
        """Converts the GGD of an IDS to VTK format.

        Args:
            time: Time step to convert. Defaults to converting the first time step.
            time_idx: Time index to convert. Defaults to converting the first time step.
            scalar_paths: A list of IDSPaths of GGD scalar arrays to convert. Defaults
                to None, in which case all scalar arrays are converted.
            vector_paths: A list of IDSPaths of GGD vector arrays to convert. Defaults
                to None, in which case all vectors arrays are converted.
            plane_config: Data class containing the interpolation settings.
            outInfo: Paraview's Source outInfo information object.
            progress: Progress indicator for Paraview.

        Returns:
            vtkPartitionedDataSetCollection containing the converted GGD data.
        """
        self.points = vtkPoints()
        self.assembly = vtkDataAssembly()
        self.time_idx = self._resolve_time_idx(time_idx, time)
        self.plane_config = plane_config

        self.progress = progress

        if self.time_idx is None:
            return None

        self.grid_ggd = get_grid_ggd(self.ids, self.time_idx, parent_idx)
        if self.grid_ggd is None:
            logger.warning("Could not load a valid GGD grid.")
            return None

        if hasattr(self.grid_ggd, "path") and self.grid_ggd.path:
            self.reference_grid_path = self.grid_ggd.path
            self._replace_grid_with_reference()
        else:
            self.reference_grid_path = None

        if not self._is_grid_valid():
            return None

        if outInfo is None:
            self.output = vtkPartitionedDataSetCollection()
        else:
            self.output = vtkPartitionedDataSetCollection.GetData(outInfo)
        self.output.SetDataAssembly(self.assembly)

        self.ps_reader = read_ps.PlasmaStateReader(self.ids)
        self.ps_reader.load_arrays_from_path(self.time_idx, scalar_paths, vector_paths)

        self.is_jorek = True if plane_config.n_plane != 0 else False
        self._fill_grid_and_plasma_state()

        return self.output

    def _replace_grid_with_reference(self):
        """
        Retrieve the GGD grid from a reference IDS path and replace the current grid.
        """
        path = self.grid_ggd.path
        logger.info(f"Fetching the GGD grid from the reference: '{path}'")
        path = path.strip("#")
        ids_name, path = path.split("/", 1)
        if ids_name not in self.dbentry.factory.ids_names():
            raise ValueError(f"{ids_name} is not a valid IDS name")

        if not self.dbentry:
            raise ValueError(
                "An opened dbentry must be provided if a GGD grid reference is given."
            )

        ids = self.dbentry.get(
            ids_name,
            autoconvert=False,
            lazy=True,
            ignore_unknown_dd_version=True,
        )
        try:
            self.grid_ggd = ids[path]
        except KeyError:
            raise ValueError(f"{path} does not exist in {ids}.")

    def _resolve_time_idx(self, time_idx, time):
        """Resolves the appropriate time index based on the given time index or time
        value.

        Args:
            time_idx: The index of the time step to convert.
            time: The time value to convert.

        Returns:
            The resolved time index, or None if an error occurs.
        """
        if time is not None and time_idx is not None:
            logger.error("The time and time index cannot be provided at the same time.")
            return None
        elif time_idx is not None:
            if time_idx >= len(self.ids.time):
                logger.error("The requested index cannot be found in the IDS.")
                return None
            return time_idx
        elif time is not None:
            indices = find_closest_indices([time], self.ids.time)
            if len(indices) == 0:
                logger.warning(
                    "No time steps found that are less than or equal to the provided "
                    "time. Converting the first time step instead."
                )
                time_idx = 0
            else:
                time_idx = indices[0]
            logger.info(
                f"Converting timestep: t = {self.ids.time[time_idx]} at index = "
                f"{time_idx}"
            )
            return time_idx
        else:
            time_idx = len(self.ids.time) // 2
            logger.info(
                "No time or time index provided, so converting the middle time "
                f"step: t = {self.ids.time[time_idx]} at index {time_idx}."
            )
            return time_idx

    def _is_grid_valid(self):
        """Validates if the grid is properly loaded."""
        if self.grid_ggd is None:
            logger.warning("Could not load a valid GGD grid.")
            return False
        if not hasattr(self.grid_ggd, "space") or len(self.grid_ggd.space) < 1:
            logger.warning("The grid_ggd does not contain a space.")
            return False
        return True

    def _fill_grid_and_plasma_state(self):
        """Fills the VTK output object with the GGD grid and GGD array values."""

        # If the grid is a reference to another grid, we store it by its path string
        # Otherwise by ID of the object
        if self.reference_grid_path:
            grid_key = str(self.reference_grid_path)
        else:
            grid_key = id(self.grid_ggd)

        ugrids = self.get_grids(grid_key, self.plane_config)
        if self.is_jorek:
            self._fill_jorek(ugrids)
        else:
            self._fill_ggd(ugrids)

    def _fill_jorek(self, ugrids):
        """Fill the GGD arrays for JOREK grid.

        Args:
            ugrids: Dictionary containing the ugrid of each subset.
        """
        n_period = self.grid_ggd.space[1].geometry_type.index
        if n_period > 0:
            read_jorek.read_plasma_state(
                self.grid_ggd, self.ps_reader, self.plane_config, ugrids[-1]
            )
            self._set_partition(0, ugrids[-1], -1)
        else:
            logger.error("Invalid plane configuration for the given IDS type.")

    def _fill_ggd(self, ugrids):
        """Fill the GGD arrays for each grid subset.

        Args:
            ugrids: Dictionary containing the ugrid of each subset.
        """
        num_subsets = len(self.grid_ggd.grid_subset)
        if num_subsets <= 1:
            logger.info("No subsets to read from grid_ggd")
            self.output.SetNumberOfPartitionedDataSets(1)
            self._set_partition(0, ugrids[-1], -1)
            self.ps_reader.read_plasma_state(-1, ugrids[-1])
        elif self.ids.metadata.name == "wall":
            # FIXME: what if num_subsets is 2 or 3?
            self.output.SetNumberOfPartitionedDataSets(num_subsets - 3)
            self._set_partition(0, ugrids[-1], -1)
            self.ps_reader.read_plasma_state(-1, ugrids[-1])

            for subset_idx in range(4, num_subsets):
                self._set_partition(subset_idx - 3, ugrids[subset_idx], subset_idx)
                self.ps_reader.read_plasma_state(subset_idx, ugrids[subset_idx])
                if self.progress:
                    self.progress.increment(0.5 / num_subsets)
        else:
            self.output.SetNumberOfPartitionedDataSets(num_subsets)
            for subset_idx in range(num_subsets):
                self._set_partition(subset_idx, ugrids[subset_idx], subset_idx)
                self.ps_reader.read_plasma_state(subset_idx, ugrids[subset_idx])
                if self.progress:
                    self.progress.increment(0.5 / num_subsets)

    def get_grids(self, grid_id, plane_config):
        """Fetches the unstructured grids at a certain time index. Note that this
        function is cached using lru_cache based on `time_idx`.

        Args:
            grid_id: ID of the GGD grid corresponding to a certain time index. This is
                used for creation of the hash of the cache entry.
            plane_config: plane_config used to generate the grid. These should be all
                zero for a regular GGD. However, for a JOREK GGD, these can be changed,
                and changing them should trigger the grid to be reloaded.

        Returns:
            Dictionary containing the ugrid of each subset, with the subset index as
            keys.
        """
        logger.info("No cache found, loading the grid from the IDS.")
        num_subsets = len(self.grid_ggd.grid_subset)
        read_geom.fill_vtk_points(
            self.grid_ggd, 0, self.points, self.ids.metadata.name, self.progress
        )
        ugrids = {}
        for subset_idx in range(-1, num_subsets):
            if subset_idx == 0 and self.progress:
                self.progress.set(0)
            progress = self.progress if subset_idx == -1 else None

            if self.is_jorek:
                ugrids[subset_idx] = (
                    read_jorek.convert_grid_subset_to_unstructured_grid(
                        self.grid_ggd, self.plane_config
                    )
                )
            else:
                ugrids[subset_idx] = (
                    read_geom.convert_grid_subset_geometry_to_unstructured_grid(
                        self.grid_ggd, subset_idx, self.points, progress
                    )
                )
            if self.progress and num_subsets != 0:
                self.progress.increment(0.5 / num_subsets)
        return ugrids

    def _set_partition(self, partition, ugrid, subset_idx):
        """Sets a partition in the output dataset and updates the assembly structure.

        Args:
            partition: Partition index in the output dataset.
            ugrid: Unstructured grid data.
            subset_idx: Index of the subset in `self.grid_ggd.grid_subset`, if negative
                the metadata name is used.
        """
        subset = None if subset_idx < 0 else self.grid_ggd.grid_subset[subset_idx]
        label = str(subset.identifier.name) if subset else self.ids.metadata.name

        self.output.SetPartition(partition, 0, ugrid)
        child = self.assembly.AddNode(label.replace(" ", "_"), 0)
        self.assembly.AddDataSetIndex(child, partition)

        self.output.GetMetaData(partition).Set(vtkCompositeDataSet.NAME(), label)

    def _write_vtk_to_xml(self, vtk_object, output_path):
        """Writes a VTK object to XML file, and a directory containing files for each
        grid subset. The directory structure looks as follows:
        .
        ├── test
        |   ├── test_0_0.vtu
        |   ├── test_1_0.vtu
        |   ├── test_2_0.vtu
        |   ├── ...
        └── test.vtpc

        Args:
            vtk_object: The VTK object to be written to disk.
            output_path: The output file path where the VTK objects will be saved.
        """
        if vtk_object is None:
            logger.error("Cannot write None object to XML.")
            return

        logger.info(f"Writing VTK file to {output_path}...")
        writer = vtkXMLPartitionedDataSetCollectionWriter()
        writer.SetInputData(vtk_object)
        writer.SetFileName(output_path.with_suffix(".vtpc"))
        writer.Write()
        logger.info(f"Successfully wrote VTK object to {output_path}")



import subprocess
from pathlib import Path

import pytest

# Load XML scripts for integration testing
ROOT_DIRECTORY = Path(__file__).resolve().parents[3]
TEST_DIRECTORY = (
    ROOT_DIRECTORY
    / "imas_paraview"
    / "tests"
    / "integration_tests"
    / "xml_integration_tests"
)
TEST_SCRIPTS = list(TEST_DIRECTORY.glob("*.xml"))


def run_test(script, timeout=120):
    """Run a single test script with ParaView and a timeout."""
    print(f"Running test script: {script}")

    try:
        result = subprocess.run(
            [
                "xvfb-run",
                "--error-file=/dev/stdout",
                "-a",
                "paraview",
                "--dr",
                # "-v",
                # "9",
                "--no-mpi",
                "--test-script",
                script,
                "--exit",
            ],
            timeout=timeout,
        )

        if result.returncode != 0:
            print(f"Test failed: {script} with return code {result.returncode}")
            return False

        return True

    except subprocess.TimeoutExpired:
        print(
            f"Test {script} timed out after {timeout} seconds. "
            "This indicates that ParaView encountered an issue and failed to complete "
            "the test, or the timeout value is too short for the test to finish."
        )
        return False
    except Exception as e:
        print(f"An error occurred: {e}")
        return False


@pytest.mark.integration
def test_xvfb_pass():
    """Test a passing integration test, by testing an empty XML."""
    test_passed = run_test(
        ROOT_DIRECTORY
        / "imas_paraview"
        / "tests"
        / "integration_tests"
        / "xml_xvfb_tests"
        / "passing_test.xml"
    )
    assert test_passed


@pytest.mark.integration
def test_xvfb_fail():
    """Test failing integration test."""
    test_passed = run_test(
        ROOT_DIRECTORY
        / "imas_paraview"
        / "tests"
        / "integration_tests"
        / "xml_xvfb_tests"
        / "failing_test.xml"
    )
    assert not test_passed


@pytest.mark.integration
@pytest.mark.parametrize(
    "test_script", [pytest.param(script, id=script.name) for script in TEST_SCRIPTS]
)
def test_integration_scripts(test_script):
    """Parameterized test function for running integration tests."""
    if "hdf5" in str(test_script) and not Path("hdf5_testdb").exists():
        pytest.skip("HDF5 testdb does not exist")
    if "mdsplus" in str(test_script) and not Path("mdsplus_testdb").exists():
        pytest.skip("MDSplus testdb does not exist")
    if "profiles_1d_mapper" in str(test_script):
        pytest.skip("Requires ITER data, not available on GH Actions")
    test_passed = run_test(test_script)
    assert test_passed



"""Plugin to view Jorek GGD data."""

import logging

from paraview.util.vtkAlgorithm import smhint, smproxy

from imas_paraview.plugins.ggd_base_reader import GGDBaseReader

logger = logging.getLogger("imas_paraview")


SUPPORTED_IDS_NAMES = [
    "mhd",
    "radiation",
    "plasma_profiles",
]


@smproxy.source(label="JOREK Reader")
@smhint.xml("""<ShowInMenu category="IMAS Tools" />""")
class JorekGGDReader(
    GGDBaseReader,
    use_bezier=True,
    is_time_dependent=True,
):
    def __init__(self):
        supported_ids = SUPPORTED_IDS_NAMES
        super().__init__(supported_ids)
        self._MIN_N_PLANE = self._n_plane = 1



from pathlib import Path

import imas
from asv_runner.benchmarks.mark import skip_benchmark

from imas_paraview.convert import Converter
from imas_paraview.tests.fill_ggd import fill_ids

SIZE_GRID = 10


def create_uri(backend):
    path = Path.cwd() / f"DB-{backend}"
    return f"imas:{backend.lower()}?path={path}"


class Convert:
    def setup(self):
        self.edge_profiles = imas.IDSFactory().edge_profiles()
        fill_ids(self.edge_profiles, grid_size=SIZE_GRID)

    def time_convert(self):
        converter = Converter(self.edge_profiles)
        converter.ggd_to_vtk(time_idx=0)


class ConvertNetCDF:
    def setup(self):
        es = imas.IDSFactory().edge_profiles()
        fill_ids(es, grid_size=SIZE_GRID)
        self.uri = Path.cwd() / "DB-netcdf.nc"
        with imas.DBEntry(self.uri, "w") as dbentry:
            dbentry.put(es)

    def time_load_and_convert(self):
        entry = imas.DBEntry(self.uri, "r")
        ids = entry.get("edge_profiles", lazy=False, autoconvert=False)
        converter = Converter(ids)
        converter.ggd_to_vtk(time_idx=0)

    def time_load_and_convert_lazy(self):
        entry = imas.DBEntry(self.uri, "r")
        ids = entry.get("edge_profiles", lazy=True, autoconvert=False)
        converter = Converter(ids)
        converter.ggd_to_vtk(time_idx=0)


# TODO: HDF5 and MDSPlus benchmarks are skipped as IMAS-Core is not yet available
class ConvertHDF5:
    def setup(self):
        es = imas.IDSFactory().edge_profiles()
        fill_ids(es, grid_size=SIZE_GRID)
        self.uri = create_uri("hdf5")
        with imas.DBEntry(self.uri, "w") as dbentry:
            dbentry.put(es)

    @skip_benchmark
    def time_load_and_convert(self):
        entry = imas.DBEntry(self.uri, "r")
        ids = entry.get("edge_profiles", lazy=False, autoconvert=False)
        converter = Converter(ids)
        converter.ggd_to_vtk(time_idx=0)

    @skip_benchmark
    def time_load_and_convert_lazy(self):
        entry = imas.DBEntry(self.uri, "r")
        ids = entry.get("edge_profiles", lazy=True, autoconvert=False)
        converter = Converter(ids)
        converter.ggd_to_vtk(time_idx=0)


class ConvertMDSPlus:
    def setup(self):
        es = imas.IDSFactory().edge_profiles()
        fill_ids(es, grid_size=SIZE_GRID)
        self.uri = create_uri("mdsplus")
        with imas.DBEntry(self.uri, "w") as dbentry:
            dbentry.put(es)

    @skip_benchmark
    def time_load_and_convert(self):
        entry = imas.DBEntry(self.uri, "r")
        ids = entry.get("edge_profiles", lazy=False, autoconvert=False)
        converter = Converter(ids)
        converter.ggd_to_vtk(time_idx=0)

    @skip_benchmark
    def time_load_and_convert_lazy(self):
        entry = imas.DBEntry(self.uri, "r")
        ids = entry.get("edge_profiles", lazy=True, autoconvert=False)
        converter = Converter(ids)
        converter.ggd_to_vtk(time_idx=0)



import logging
from unittest.mock import Mock, patch

import pytest

from imas_paraview import VTKHandler


@pytest.fixture
def mock_output_window():
    with patch("imas_paraview.vtkhandler.win") as mock_vtk_win:
        mock_output_window = Mock()
        mock_vtk_win.GetInstance.return_value = mock_output_window
        yield mock_output_window


@pytest.fixture
def logger(request):
    # Ensure each test gets its own logger
    logger_name = f"test_logger_{request.node.name}"
    logger = logging.getLogger(logger_name)
    handler = VTKHandler()
    logger.addHandler(handler)
    return logger


def test_error_log(mock_output_window, logger):
    """Test logging behavior across different levels."""

    logger.setLevel(logging.ERROR)
    logger.error("This is an error message")

    # Ensure other text types were not called
    mock_output_window.DisplayWarningText.assert_not_called()
    mock_output_window.DisplayDebugText.assert_not_called()
    mock_output_window.DisplayText.assert_not_called()

    # Ensure the error text was called
    mock_output_window.DisplayErrorText.assert_called_once()


def test_warning_log(mock_output_window, logger):
    """Test logging behavior for WARNING level."""

    logger.setLevel(logging.WARNING)
    logger.warning("This is a warning message")

    # Ensure other text types were not called
    mock_output_window.DisplayErrorText.assert_not_called()
    mock_output_window.DisplayDebugText.assert_not_called()
    mock_output_window.DisplayText.assert_not_called()

    # Ensure the warning text was called
    mock_output_window.DisplayWarningText.assert_called_once()


def test_info_log(mock_output_window, logger):
    """Test logging behavior for INFO level."""

    logger.setLevel(logging.INFO)
    logger.info("This is an info message")

    # Ensure other text types were not called
    mock_output_window.DisplayErrorText.assert_not_called()
    mock_output_window.DisplayWarningText.assert_not_called()
    mock_output_window.DisplayDebugText.assert_not_called()

    # Ensure the general text was called
    mock_output_window.DisplayText.assert_called_once()


def test_debug_log(mock_output_window, logger):
    """Test logging behavior for DEBUG level."""

    logger.setLevel(logging.DEBUG)
    logger.debug("This is a debug message")

    # Ensure other text types were not called
    mock_output_window.DisplayErrorText.assert_not_called()
    mock_output_window.DisplayWarningText.assert_not_called()
    mock_output_window.DisplayText.assert_not_called()

    # Ensure the debug text was called
    mock_output_window.DisplayDebugText.assert_called_once()



import logging
import sys
from collections import OrderedDict
from pathlib import Path

import click
import imas
import imas.backends.imas_core.imas_interface
from imas.backends.imas_core.imas_interface import ll_interface
from rich import box, console, traceback
from rich.table import Table

import imas_paraview
from imas_paraview.convert import Converter
from imas_paraview.util import find_closest_indices

logger = logging.getLogger(__name__)


def _excepthook(type_, value, tb):
    logger.debug("Suppressed traceback:", exc_info=(type_, value, tb))
    # Only display the last traceback frame:
    if tb is not None:
        while tb.tb_next:
            tb = tb.tb_next
    rich_tb = traceback.Traceback.from_exception(type_, value, tb, extra_lines=0)
    console.Console(stderr=True).print(rich_tb)


@click.group("imas-paraview", invoke_without_command=True, no_args_is_help=True)
@click.option("-v", "--version", is_flag=True, help="Show version information")
def cli(version):
    """IMAS-ParaView command line interface.

    Please use one of the available commands listed below. You can get help for each
    command by executing:

        imas-paraview <command> --help
    """
    # Limit the traceback to 1 item: avoid scaring CLI users with long traceback prints
    # and let them focus on the actual error message
    sys.excepthook = _excepthook

    if version:
        print_version()


def print_version():
    """Print version information of IMAS-ParaView."""
    cons = console.Console()
    grid = Table(
        title="IMAS-ParaView version info", show_header=False, title_style="bold"
    )
    grid.box = box.HORIZONTALS
    if cons.size.width > 120:
        grid.width = 120
    grid.add_row("IMAS-ParaView version:", imas_paraview.__version__)
    grid.add_section()
    grid.add_row("IMAS-Python version:", imas.__version__)
    grid.add_section()
    grid.add_row("Default data dictionary version:", imas.IDSFactory().dd_version)
    dd_versions = ", ".join(imas.dd_zip.dd_xml_versions())
    grid.add_row("Available data dictionary versions:", dd_versions)
    grid.add_section()
    grid.add_row("Access Layer core version:", ll_interface.get_al_version() or "N/A")
    console.Console().print(grid)


@cli.command("ggd2vtk")
@click.argument("uri", type=str)
@click.argument("output_dir", type=Path)
@click.option(
    "--index",
    "-i",
    type=str,
    help="Specify an index, a list of indices or a range of indices to convert.",
)
@click.option(
    "--time",
    "-t",
    type=str,
    help="Specify a time step, a list of time steps or a time step range in seconds to "
    "convert.",
)
@click.option(
    "--all-times", "-a", is_flag=True, help="Convert all available time steps."
)
@click.option(
    "--format",
    "-f",
    type=click.Choice(["xml", "vtkhdf"], case_sensitive=True),
    default="xml",
    help="VTK output format: 'xml' (default) for standard VTK XML files, "
    "or 'vtkhdf' for VTK files using HDF5 format.",
)
@click.option("--lazy", is_flag=True, help="Enable lazy loading.")
@click.option("--no-lazy", is_flag=True, help="Disable lazy loading.")
def convert_ggd_to_vtk(
    uri,
    output_dir,
    index,
    time,
    all_times,
    format,
    lazy,
    no_lazy,
):
    """
    Convert a GGD structure in an IDS and write the converted VTK file to disk.
    Optionally, specific time values or time indices can be specified to be converted.

    \b
    Arguments:
    \b
    uri             URI of the Data Entry. This should contain the IDS and its fragment.
    output_dir      Output directory to store the VTK files.

    ------------------------ Examples ------------------------

    Without time slicing:

     \b
     To convert the middle time step (default), no time options need to be given:
         $ ggd2vtk imas:hdf5?path=testdb#edge_profiles test_dir
     \b
     To convert all time slices in the IDS:
         $ ggd2vtk imas:hdf5?path=testdb#edge_profiles test_dir -a
     \b
     To convert occurrence number 1 (default=0):
         $ ggd2vtk imas:hdf5?path=testdb#edge_profiles:1 test_dir

    Index-based slicing:

     \b
     To convert index 5:
         $ ggd2vtk imas:hdf5?path=testdb#edge_profiles test_dir -i 5
     \b
     To convert indices 5, 8, and 9:
         $ ggd2vtk imas:hdf5?path=testdb#edge_profiles test_dir -i 5,8,9
     \b
     To convert a range of indices, such as 2,3,4,5,6:
         $ ggd2vtk imas:hdf5?path=testdb#edge_profiles test_dir -i 2:6

    Time-based slicing:

     \b
     To convert time step at 5.5s:
         $ ggd2vtk imas:hdf5?path=testdb#edge_profiles test_dir -t 5.5
     \b
     To convert time steps 5.5s, 8s, and 9.1s:
         $ ggd2vtk imas:hdf5?path=testdb#edge_profiles test_dir -t 5.5,8,9.1
     \b
     To convert all time steps that fall between 2.2s and 6.6s:
         $ ggd2vtk imas:hdf5?path=testdb#edge_profiles test_dir -t 2.2:6.6

        \b
        Note: If the specified time step is not found in the IDS, the closest earlier
        time step will be used by default.
    """

    sys.excepthook = _excepthook
    uri, ids_name, occurrence = parse_uri(uri)
    click.echo(f"Loading {ids_name} from {uri} with occurrence {occurrence}...")
    with imas.DBEntry(uri, "r") as entry:
        click.echo("Loading IDS...")
        use_lazy = is_lazy(all_times, lazy, no_lazy)
        ids = entry.get(
            ids_name,
            lazy=use_lazy,
            occurrence=occurrence,
            autoconvert=False,
        )
        index_list = parse_time_options(ids.time, index, time, all_times)

        click.echo("Converting GGD to a VTK file...")

        # TODO: Add time-dependent VTKHDF conversion
        if format == "xml":
            converter = Converter(ids, dbentry=entry)
            converter.write_to_xml(output_dir, index_list)

        elif format == "vtkhdf":
            raise NotImplementedError("vtkhdf format is not yet implemented.")


def is_lazy(all_times, lazy_flag, no_lazy_flag):
    """Determine whether to enable or disable lazy loading, based on
    if the all_times flag is enabled. Optionally, the default behaviour can be
    overridden by using the --lazy or --no-lazy flag.
    Args:
        all_times: Flag to convert all time steps in an IDS.
        lazy_flag: Flag to enable lazy loading.
        no_lazy_flag: Flag to disable lazy loading.
    """
    if lazy_flag and no_lazy_flag:
        click.echo("Both --lazy and --no-lazy flag were provided. Ignoring...")
        lazy_flag = None
        no_lazy_flag = None

    # Lazy loading is enabled by default
    use_lazy_loading = True
    if all_times:
        use_lazy_loading = False

    # Override default if flag is provided
    if lazy_flag:
        use_lazy_loading = True
    elif no_lazy_flag:
        use_lazy_loading = False

    if use_lazy_loading:
        click.echo("Lazy loading is enabled.")
    else:
        click.echo("Lazy loading is disabled.")
    return use_lazy_loading


def parse_uri(uri):
    """Parses the URI according to the IMAS URI Scheme Documentation v0.3.
    Parses the URI and extracts the fragment part of the URI.

    Example:
        Given the URI:
            uri = "imas:hdf5?path=testdb#edge_profiles:1"

        Returns:
            uri_no_frag = "imas:hdf5?path=testdb"
            ids_name = "edge_profiles"
            occurrence = 1

    Args:
        uri: URI to parse, should contain a fragment denoting the IDS name.

    Returns:
        uri_no_frag: URI without fragment part
        ids_name: Name of the IDS.
        occurrence: Occurrence number of the IDS.
    """
    if "#" in uri:
        split_uri = uri.split("#")
        uri_no_frag = split_uri[0]
        fragment = split_uri[1]
        if "/" in fragment:
            raise click.UsageError(
                "It is currently not possible to select an IDS subset for conversion."
                "It is only possible to convert the entire IDS."
            )
        elif ":" in fragment:
            split_fragment = fragment.split(":")
            ids_name = split_fragment[0]
            occurrence = int(split_fragment[1])
        else:
            ids_name = fragment
            occurrence = 0
    else:
        raise click.UsageError(
            "The IDS must be provided as a fragment to the URI. For example: "
            'uri = "imas:hdf5?path=testdb#edge_profiles"'
        )
    return uri_no_frag, ids_name, occurrence


def parse_time_options(ids_time, index, time, all_times):
    """Parses the provided time options. Only a single time option (index, time or
    all_times) may be used. Either a single value ("5"), a list of values ("2,3,4") or
    a range of values ("2:4") may be provided for the index or time parameters.

    Args:
        index: String containing either a single index, a list of indices or a range of
            indices. Alternatively, it can be None.
        time: String containing either a single time step, a list of time steps or a
            range of time steps. Alternatively, it can be None.
        all_times: Boolean to convert all time steps

    Returns:
        A list containing the time indices to convert.
    """
    # Check if more than a single time option is provided
    if sum([index is not None, time is not None, all_times]) > 1:
        raise click.UsageError(
            "You can only provide one time-related argument: either --time, --index, "
            "or --all-times."
        )
    if index:
        index_list = parse_index(index)
    elif time:
        index_list = parse_time(ids_time, time)
    elif all_times:
        index_list = list(range(len(ids_time)))
        click.echo(
            "Converting all time steps in the IDS. Depending on the number of time "
            f"steps, this could take a while. Converting a total of {len(index_list)} "
            "time steps."
        )
    else:
        index = len(ids_time) // 2
        middle_time = ids_time[index]
        click.echo(
            "No time options were set, so only converting the middle time step: "
            f"t = {middle_time}  at index {index}"
        )
        index_list = [index]

    if index_list == []:
        raise click.UsageError(
            "Could not find any time steps for the provided time steps."
        )
    return index_list


def parse_index(index):
    """Parses the index parameter and returns a list of IDS time indices to convert.
    Either a single integer ("5"), a list of integers ("2,3,4") or a range of integers
    ("2:4") may be provided.

    Args:
        index: String containing either a single index, a list of indices or a range of
            indices.

    Returns:
        A list containing the time indices to convert.
    """
    # Single index
    if index.isdigit():
        index_list = [int(index)]
    # List of indices
    elif "," in index:
        for input_index in index.split(","):
            if not input_index.strip().isdigit():
                raise click.UsageError("All indices in given list must be integers.")
        index_list = [int(x.strip()) for x in index.split(",")]
        indices_dict = OrderedDict.fromkeys(index_list)
        if len(index_list) != len(indices_dict):
            click.echo(
                "Duplicate time step indices were detected. Any duplicate indices will "
                "be ignored."
            )

            index_list = list(indices_dict)
    # Range of indices
    elif ":" in index:
        if index.count(":") > 1:
            raise click.UsageError("Only a single range may be provided.")
        start_str, end_str = index.split(":")
        if not start_str.strip().isdigit() or not end_str.strip().isdigit():
            raise click.UsageError(
                "The lower and upper bound of indices must be integers."
            )
        start = int(start_str)
        end = int(end_str)
        if end < start:
            raise click.UsageError(
                "The final time index in range must be greater than the first."
            )
        index_list = list(range(start, end + 1))
    else:
        raise click.UsageError(
            "Could not determine which indices should be converted.\n"
            "Provide either a single integer ('-i 5'), "
            "a list of integers ('-i 2,3,4') or "
            "a range of integers ('-i 2:4')"
        )
    click.echo(f"Converting the following indices: {index_list}")
    return index_list


def parse_time(ids_times, time):
    """Parses the time parameter and returns a list of IDS time indices to convert.
    Either a single float ("5.0"), a list of integers ("2.0,3,4.4") or a range of
    floats ("2.3:4.3") may be provided. If the specified time step is not found in the
    IDS, the closest earlier time step will be converted instead.

    Args:
        time: String containing either a single time step, a list of time steps or a
            range of time steps.

    Returns:
        A list containing the time indices to convert.
    """
    # List of time steps
    if "," in time:
        for input_time in time.split(","):
            try:
                float(input_time)
            except ValueError:
                raise click.UsageError(
                    "All time steps in given list must be valid floats."
                )
        time_list = [float(x.strip()) for x in time.split(",")]
        index_list = find_closest_indices(time_list, ids_times)
        indices_dict = OrderedDict.fromkeys(index_list)
        if len(index_list) != len(indices_dict):
            click.echo(
                "Duplicate time steps were detected. Note that provided time steps "
                "will be rounded down to a time in the IDS, and the duplicate time "
                "steps will be ignored. For example, if the IDS contains times "
                "[1.0, 2.0] and the provided time steps are t=1.1,1.8 both are rounded "
                "down to time step 1.0, and as a result only the time step at t=1.0 is "
                "converted."
            )
            index_list = list(indices_dict)
    # Range of time steps
    elif ":" in time:
        if time.count(":") > 1:
            raise click.UsageError("Only a single range may be provided.")
        start_str, end_str = time.split(":")
        try:
            start = float(start_str)
            end = float(end_str)
        except ValueError:
            raise click.UsageError(
                "The minimum and maximum range values must be valid floats."
            )
        if end < start:
            raise click.UsageError(
                "The final time index in range must be greater than the first."
            )
        index_list = [
            index for index, value in enumerate(ids_times) if start <= value <= end
        ]
    # Single time step
    else:
        try:
            time_list = [float(time)]
            index_list = find_closest_indices(time_list, ids_times)
        except ValueError:
            raise click.UsageError(
                "Could not determine which time steps should be converted.\n"
                "Provide either a single float ('-t 5.0'), "
                "a list of floats ('-t 2.1,3.5,4') or "
                "a range of floats ('-t 2.2:4.4')"
            )
    click.echo(f"Converting the following time steps: {ids_times[index_list]}")
    return index_list


if __name__ == "__main__":
    cli()



"""Plugin to view limiter structures of 2d wall descriptions nodes"""

import logging
from dataclasses import dataclass

from imas.ids_structure import IDSStructure
from paraview.util.vtkAlgorithm import smhint, smproxy
from vtkmodules.vtkCommonDataModel import vtkMultiBlockDataSet

from imas_paraview.ids_util import get_object_by_name
from imas_paraview.plugins.base_class import GGDVTKPluginBase
from imas_paraview.util import points_to_vtkpoly

logger = logging.getLogger("imas_paraview")


@dataclass
class Limiter:
    """Data class that stores limiter units, along with its name."""

    name: str
    unit: IDSStructure


@smproxy.source(label="Wall Limiter Reader")
@smhint.xml("""<ShowInMenu category="IMAS Tools" />""")
class WallLimiterReader(GGDVTKPluginBase):
    """IMAS-Python based reader that can load the limiter structures of 2D wall
    descriptions"""

    def __init__(self):
        super().__init__("vtkMultiBlockDataSet", ["wall"])

    def GetAttributeArrayName(self, idx) -> str:
        return self._selectable[idx].name

    def RequestData(self, request, inInfo, outInfo):
        if self._dbentry is None or not self._ids_and_occurrence or self._ids is None:
            return 1

        if len(self._selected) > 0:
            output = vtkMultiBlockDataSet.GetData(outInfo)
            self._load_limiters(output)
        return 1

    def setup_ids(self):
        """
        Select which limiters to show in the array domain selector. The description_2d
        AoS is searched through for limiter structures. Their names are generated based
        on the limiter type name and the limiter unit name, and added to the array
        domain selector.
        """
        assert self._ids is not None, "IDS cannot be empty during setup."

        if not self._ids.description_2d.has_value:
            raise ValueError("This IDS does not contain a description_2d node")

        descriptions = self._ids.description_2d
        self._selectable = []

        for i, description in enumerate(descriptions):
            description_name = description.type.name
            if description_name == "":
                description_name = f"description {i}"
                logger.warning(
                    "Found a limiter without a description name, "
                    f"it will be loaded as {description_name}"
                )

            limiter = description.limiter
            type_name = limiter.type.name
            if type_name == "":
                type_name = f"type {i}"
                logger.warning(
                    "Found a limiter without a type name, "
                    f"it will be loaded as {type_name}"
                )

            for j, unit in enumerate(limiter.unit):
                unit_name = unit.name
                if unit_name == "":
                    unit_name = f"unit {j}"
                    logger.warning(
                        "Found a limiter unit without a name, "
                        f"it will be loaded as {unit_name}"
                    )
                name = " / ".join(
                    [str(description_name), str(type_name), str(unit_name)]
                )
                selectable = Limiter(name, unit)
                self._selectable.append(selectable)

    def _load_limiters(self, output):
        """Go through the list of selected limiters, and load each of them in a
        separate vtkPolyData object, which are all combined into a vtkMultiBlockDataSet.

        Args:
            output: The vtkMultiBlockDataSet containing the limiter contours.
        """
        for i, limiter_name in enumerate(self._selected):
            limiter = get_object_by_name(self._selectable, limiter_name)
            if limiter is None:
                raise ValueError(f"Could not find {limiter_name}")

            logger.info(f"Selected {limiter.name}")
            vtk_poly = self._create_contour(limiter)
            output.SetBlock(i, vtk_poly)

    def _create_contour(self, limiter):
        """Create a contour based on the r,z coordinates in the limiter.
        The r,z-coordinates are stored as vtkPoints, and connected using vtkLines, which
        are both stored in a vtkPolyData object. If the contour is closed, the start
        and end points are connected.

        Args:
            limiter: limiter object containing contour

        Returns:
            vtkPolyData containing contour data
        """
        # closed was removed in DD4 - data providers need to repeat the first point
        # for closed outlines, which Just Works with our is_closed=False logic
        if getattr(limiter.unit, "closed", 0) == 0:
            is_closed = False
        else:
            is_closed = True

        r = limiter.unit.outline.r
        z = limiter.unit.outline.z
        points = [(ri, 0.0, zi) for ri, zi in zip(r, z)]

        assert len(r) == len(z), "r and z must have the same length."

        vtk_poly = points_to_vtkpoly(points, is_closed)
        return vtk_poly



import numpy as np
import pytest
from imas import DBEntry
from vtk.util.numpy_support import vtk_to_numpy
from vtkmodules.vtkCommonDataModel import vtkMultiBlockDataSet

from imas_paraview.plugins.wall_limiter import WallLimiterReader


@pytest.mark.skip(reason="no IMAS-Core available")
def test_load_limiters():
    """Test if limiters are loaded in the VTK Multiblock Dataset."""
    reader = WallLimiterReader()

    with DBEntry(
        "imas:hdf5?path=/work/imas/shared/imasdb/ITER_MACHINE_DESCRIPTION/3/116000/5/",
        "r",
    ) as entry:
        ids = entry.get("wall", lazy=True, autoconvert=False)
        reader._ids = ids
        reader.setup_ids()
        description_name = ids.description_2d[0].type.name
        limiter_name = ids.description_2d[0].limiter.type.name

        unit1 = ids.description_2d[0].limiter.unit[0]
        name1 = f"{description_name} / {limiter_name} / {unit1.name}"
        unit2 = ids.description_2d[0].limiter.unit[1]
        name2 = f"{description_name} / {limiter_name} / {unit2.name}"

        # 1 selection
        output = vtkMultiBlockDataSet()
        reader._selected = [name1]
        reader._load_limiters(output)
        assert output.GetNumberOfBlocks() == 1
        assert_values_match(unit1, output.GetBlock(0))

        # 2 selections
        output = vtkMultiBlockDataSet()
        reader._selected = [name1, name2]
        reader._load_limiters(output)
        assert output.GetNumberOfBlocks() == 2
        assert_values_match(unit1, output.GetBlock(0))
        assert_values_match(unit2, output.GetBlock(1))


def assert_values_match(unit, block):
    """Check if values in limiter unit match with the values in vtk block"""
    vtk_points = block.GetPoints()
    numpy_points = vtk_to_numpy(vtk_points.GetData())
    assert np.all(np.isclose(unit.outline.r, numpy_points[:, 0]))
    assert np.all(np.isclose(unit.outline.z, numpy_points[:, 2]))



"""
Error handling for imas open/create or ids related get/put
"""


class InvalidIDSIOError(Exception):
    def __init__(self, msg, ids_func, *ids_args):
        self.message = (
            f"Error message: {msg} | function {ids_func}, args: {list(ids_args)}"
        )

    def __str__(self):
        return self.message


class InvalidIDSParametersError(Exception):
    def __init__(self, **params):
        self.message = (
            f"Invalid IDS with parameters {params if len(params) else 'None'}"
        )

    def __str__(self):
        return self.message


class InputPointsError(Exception):
    def __init__(self, *args, **kwargs):
        pass



import imas
import pytest
from imas.ids_defs import IDS_TIME_MODE_HOMOGENEOUS
from imas.ids_path import IDSPath

from imas_paraview.convert import Converter
from imas_paraview.io.read_ps import PlasmaStateReader
from imas_paraview.tests.fill_ggd import fill_ids
from imas_paraview.util import get_grid_ggd


def test_ggd_to_vtk(dummy_ids):
    """Test if ggd_to_vtk converts all GGD arrays to VTK arrays."""
    ps = PlasmaStateReader(dummy_ids)
    scalar_paths, vector_paths, _, _ = ps.load_paths_from_ids(return_empty=True)
    ggd_names = names_from_ids(dummy_ids, scalar_paths, vector_paths)

    # Check if names of VTK object match the GGD array names
    converter = Converter(dummy_ids)
    vtk_object = converter.ggd_to_vtk()

    vtk_array_names = names_from_vtk(vtk_object)
    assert vtk_array_names == ggd_names


def test_ggd_to_vtk_reference(tmp_path):
    """Test if ggd_to_vtk works with a reference grid."""

    with imas.DBEntry(f"{tmp_path}/test.nc", "w") as dbentry:
        ids = dbentry.factory.new("edge_profiles")
        fill_ids(ids)
        dbentry.put(ids)
        ids2 = dbentry.factory.new("edge_sources")
        ids2.ids_properties.homogeneous_time = IDS_TIME_MODE_HOMOGENEOUS
        ids2.grid_ggd.resize(1)
        ids2.time = [0]

        converter = Converter(ids2, dbentry=dbentry)
        assert converter.ggd_to_vtk() is None

        ids2.grid_ggd[0].path = "#edge_profiles/grid_ggd(1)"
        converter = Converter(ids2, dbentry=dbentry)
        assert converter.ggd_to_vtk() is not None


def test_ggd_to_vtk_index(dummy_ids_five_steps):
    """Test if ggd_to_vtk works with different time indices."""
    ps = PlasmaStateReader(dummy_ids_five_steps)
    scalar_paths, vector_paths, _, _ = ps.load_paths_from_ids(return_empty=True)
    ggd_names = names_from_ids(dummy_ids_five_steps, scalar_paths, vector_paths)

    converter = Converter(dummy_ids_five_steps)
    # Check if names of VTK object match the GGD array names
    for time_idx in range(5):
        vtk_object = converter.ggd_to_vtk(time_idx=time_idx)
        vtk_array_names = names_from_vtk(vtk_object)
        assert vtk_array_names == ggd_names


def test_ggd_to_vtk_time(dummy_ids_five_steps):
    """Test if ggd_to_vtk works with different times."""
    ps = PlasmaStateReader(dummy_ids_five_steps)
    scalar_paths, vector_paths, _, _ = ps.load_paths_from_ids(return_empty=True)
    ggd_names = names_from_ids(dummy_ids_five_steps, scalar_paths, vector_paths)

    converter = Converter(dummy_ids_five_steps)
    # Check if names of VTK object match the GGD array names
    for time in range(5):
        vtk_object = converter.ggd_to_vtk(time=time)
        vtk_array_names = names_from_vtk(vtk_object)
        assert vtk_array_names == ggd_names


def test_ggd_to_vtk_out_of_bounds(dummy_ids_five_steps):
    """Test if ggd_to_vtk fails when given an index which is not in the IDS."""
    time_idx = 6
    converter = Converter(dummy_ids_five_steps)
    vtk_object = converter.ggd_to_vtk(time_idx=time_idx)
    assert vtk_object is None


def test_ggd_to_vtk_subset():
    """Test for conversion of a subset of GGD arrays."""
    ids = imas.IDSFactory(version="3.41.0").new("edge_profiles")
    fill_ids(ids)

    # Convert subset of filled paths
    es_scalar_paths = [
        IDSPath("ggd/ion/state/density"),
        IDSPath("ggd/electrons/temperature"),
        IDSPath("ggd/phi_potential"),
    ]
    es_vector_paths = [
        IDSPath("ggd/neutral/state/velocity"),
        IDSPath("ggd/ion/velocity"),
        IDSPath("ggd/e_field"),
    ]
    converter = Converter(ids)
    vtk_object = converter.ggd_to_vtk(
        scalar_paths=es_scalar_paths, vector_paths=es_vector_paths
    )
    vtk_array_names = names_from_vtk(vtk_object)
    ggd_names = names_from_ids(ids, es_scalar_paths, es_vector_paths)
    assert vtk_array_names == ggd_names


def test_ggd_to_vtk_subset_time_index(dummy_ids_five_steps):
    """Test if ggd_to_vtk returns None when given time and index values."""
    converter = Converter(dummy_ids_five_steps)
    vtk_object = converter.ggd_to_vtk(time=5, time_idx=6)
    assert vtk_object is None


def assert_cache(converter, hits, misses):
    """Assert that the cache of a converter has a certain number of hits and misses."""
    cache = converter.get_grids.cache_info()
    assert cache.hits == hits
    assert cache.misses == misses


def test_ggd_to_vtk_grid_caching(dummy_ids):
    """Test if ggd_to_vtk caches grids if the same time step is provided."""
    converter = Converter(dummy_ids)
    vtk_object1 = converter.ggd_to_vtk()
    assert_cache(converter, 0, 1)
    vtk_object2 = converter.ggd_to_vtk()
    assert_cache(converter, 1, 1)
    vtk_object3 = converter.ggd_to_vtk()
    assert_cache(converter, 2, 1)

    assert (
        names_from_vtk(vtk_object1)
        == names_from_vtk(vtk_object2)
        == names_from_vtk(vtk_object3)
    )


def test_ggd_to_vtk_grid_caching_time_dependent(dummy_ids_five_steps):
    """Test if ggd_to_vtk caches grids if the same time step is provided."""
    converter = Converter(dummy_ids_five_steps)
    for time_idx in range(5):
        _ = converter.ggd_to_vtk(time_idx=time_idx)
        assert_cache(converter, 0, time_idx + 1)

    for time_idx in range(5):
        _ = converter.ggd_to_vtk(time_idx=time_idx)
        assert_cache(converter, time_idx + 1, 5)


def test_convert_to_xml(dummy_ids_five_steps, tmp_path):
    """Test if convert_to_xml converts a single index."""
    output_file = tmp_path / "test"
    converter = Converter(dummy_ids_five_steps)
    converter.write_to_xml(output_file)
    assert_output_exists(dummy_ids_five_steps, 0, output_file)


def test_convert_to_xml_out_of_bounds(dummy_ids_five_steps, tmp_path):
    """Test if convert_to_xml fails when given an index which is not in the IDS."""
    output_file = tmp_path / "test_.vtpc"
    converter = Converter(dummy_ids_five_steps)
    with pytest.raises(RuntimeError):
        converter.write_to_xml(output_file, [6])


def test_convert_to_xml_index(dummy_ids_five_steps, tmp_path):
    """Test if convert_to_xml converts a single index."""
    converter = Converter(dummy_ids_five_steps)

    for time_idx in range(5):
        output_file = tmp_path / f"test_{time_idx}"
        converter.write_to_xml(output_file, [time_idx])
        assert_output_exists(dummy_ids_five_steps, time_idx, output_file)


def test_convert_to_xml_index_list(dummy_ids_five_steps, tmp_path):
    """Test if convert_to_xml converts an index list."""

    output_file = tmp_path / "test"
    time_idx = [0, 1, 2, 3, 4]
    converter = Converter(dummy_ids_five_steps)
    converter.write_to_xml(output_file, time_idx)
    for time_idx in range(5):
        assert_output_exists(dummy_ids_five_steps, time_idx, output_file)


def assert_output_exists(ids, time_idx, output_file):
    """Assert the VTK object is correctly written to disk."""
    output_dir = output_file.parent / output_file.stem

    # Check if vtpc file and the directory exists
    assert output_file.exists()
    assert output_dir.is_dir()

    # Check if the vtu files exist
    grid_ggd = get_grid_ggd(ids, time_idx)
    num_subsets = len(grid_ggd.grid_subset)
    for n in range(num_subsets):
        vtu_file = output_file.stem + f"_{n}_0.vtu"
        assert output_dir / vtu_file


def names_from_ids(ids, scalar_paths, vector_paths):
    """Convert the names of the GGD array paths to the names given in Paraview."""
    ps = PlasmaStateReader(ids)
    ps.load_arrays_from_path(0, scalar_paths, vector_paths)
    ggd_names = set()
    for array in ps.scalar_array_list + ps.vector_array_list:
        name = ps._create_name_with_units(array)
        ggd_names.add(name)
    return ggd_names


def names_from_vtk(vtk_object):
    """Extract the array names from the VTK object."""
    n_partds = vtk_object.GetNumberOfPartitionedDataSets()
    array_names = set()
    for i in range(n_partds):
        part_ds = vtk_object.GetPartitionedDataSet(i)
        n_part = part_ds.GetNumberOfPartitions()
        for j in range(n_part):
            part = part_ds.GetPartition(j)
            cell_data = part.GetCellData()
            num_arrays = cell_data.GetNumberOfArrays()
            for k in range(num_arrays):
                array_name = cell_data.GetArrayName(k)
                array_names.add(array_name)
    return array_names



import imas
import pytest
from imas.ids_path import IDSPath
from vtkmodules.vtkCommonCore import vtkPoints

from imas_paraview.io.read_geom import (
    convert_grid_subset_geometry_to_unstructured_grid,
    fill_vtk_points,
)
from imas_paraview.io.read_ps import PlasmaStateReader
from imas_paraview.tests.fill_ggd import fill_scalar_quantity, fill_vector_quantity
from imas_paraview.util import get_grid_ggd


def test_read_plasma_state(ids_name, dummy_ids):
    """Tests reading plasma state from the IDS."""
    if ids_name == "transport_solver_numerics":
        pytest.skip(reason="boundary_conditions_ggd are not filled yet")
    elif ids_name == "runaway_electrons":
        pytest.skip(reason="runaway_electrons is not implemented")

    subset_idx = 0
    space_idx = 0
    grid_ggd = get_grid_ggd(dummy_ids)
    points = vtkPoints()
    fill_vtk_points(grid_ggd, space_idx, points, ids_name)
    ugrid = convert_grid_subset_geometry_to_unstructured_grid(
        grid_ggd, subset_idx, points
    )
    ps_reader = PlasmaStateReader(dummy_ids)
    ps_reader.read_plasma_state(subset_idx, ugrid)


def test_create_name_with_units():
    ids = imas.IDSFactory(version="3.40.0").new("edge_sources")
    ps_reader = PlasmaStateReader(ids)

    # Resize relevant structures
    ids.source.resize(1)
    ids.source[0].ggd.resize(1)
    ids.source[0].ggd[0].ion.resize(1)
    ids.source[0].ggd[0].ion[0].state.resize(1)
    array = ids.source[0].ggd[0].ion[0].state[0].energy

    # Set identifier name and labels
    ids.source[0].identifier.name = "Charge exchange"
    ids.source[0].ggd[0].ion[0].label = "Ar"
    ids.source[0].ggd[0].ion[0].state[0].label = "Ar+14"

    name = ps_reader._create_name_with_units(array)
    assert name == "Source (Charge exchange) Ion (Ar) State (Ar+14) Energy [W.m^-3]"


def test_load_paths_from_ids_empty():
    """Test if load_paths_from_ids returns empty lists if no GGDs are filled"""
    ids = imas.IDSFactory(version="3.41.0").new("edge_sources")
    ps_reader = PlasmaStateReader(ids)

    _, _, scalar_array_paths, vector_array_paths = ps_reader.load_paths_from_ids()
    assert scalar_array_paths == vector_array_paths == []
    ids.source.resize(1)
    _, _, scalar_array_paths, vector_array_paths = ps_reader.load_paths_from_ids()
    assert scalar_array_paths == vector_array_paths == []


def test_load_paths_from_ids_filled():
    """Test if load_paths_from_ids returns filled GGD paths"""
    ids = imas.IDSFactory(version="3.41.0").new("edge_sources")
    ps_reader = PlasmaStateReader(ids)

    ids.source.resize(1)
    ids.source[0].ggd.resize(1)
    ids.source[0].ggd[0].ion.resize(1)
    ids.source[0].ggd[0].ion[0].state.resize(1)
    energy = ids.source[0].ggd[0].ion[0].state[0].energy
    momentum = ids.source[0].ggd[0].ion[0].state[0].momentum
    _, _, scalar_array_paths, vector_array_paths = ps_reader.load_paths_from_ids()
    assert scalar_array_paths == vector_array_paths == []

    fill_scalar_quantity(energy, 1, 1, 1)
    fill_vector_quantity(momentum, 1, 1, 1)
    _, _, scalar_array_paths, vector_array_paths = ps_reader.load_paths_from_ids()
    assert scalar_array_paths == [energy.metadata.path]
    assert vector_array_paths == [momentum.metadata.path]


def test_load_paths_from_ids_empty_first():
    """Test if load_paths_from_ids handles first entry in IDSStructure being empty"""
    ids = imas.IDSFactory(version="3.41.0").new("edge_sources")
    ps_reader = PlasmaStateReader(ids)

    ids.source.resize(1)
    ids.source[0].ggd.resize(1)
    ids.source[0].ggd[0].ion.resize(2)
    ids.source[0].ggd[0].ion[1].state.resize(1)
    energy = ids.source[0].ggd[0].ion[1].state[0].energy
    momentum = ids.source[0].ggd[0].ion[1].state[0].momentum

    _, _, scalar_array_paths, vector_array_paths = ps_reader.load_paths_from_ids()
    assert scalar_array_paths == vector_array_paths == []

    fill_scalar_quantity(energy, 1, 1, 1)
    fill_vector_quantity(momentum, 1, 1, 1)
    _, _, scalar_array_paths, vector_array_paths = ps_reader.load_paths_from_ids()
    assert scalar_array_paths == [energy.metadata.path]
    assert vector_array_paths == [momentum.metadata.path]


def test_load_paths_from_ids_all():
    """Test if load_paths_from_ids returns all GGD paths with return_empty flag"""
    ids = imas.IDSFactory(version="3.41.0").new("edge_sources")
    ps_reader = PlasmaStateReader(ids)

    # Only load filled GGD array paths
    scalar_array_paths, vector_array_paths, _, _ = ps_reader.load_paths_from_ids(
        return_empty=True
    )

    es_scalar_v3_41 = [
        IDSPath("source/ggd/electrons/particles"),
        IDSPath("source/ggd/electrons/energy"),
        IDSPath("source/ggd/ion/energy"),
        IDSPath("source/ggd/ion/particles"),
        IDSPath("source/ggd/ion/state/particles"),
        IDSPath("source/ggd/ion/state/energy"),
        IDSPath("source/ggd/neutral/particles"),
        IDSPath("source/ggd/neutral/energy"),
        IDSPath("source/ggd/neutral/state/particles"),
        IDSPath("source/ggd/neutral/state/energy"),
        IDSPath("source/ggd/total_ion_energy"),
        IDSPath("source/ggd/current"),
    ]
    es_vector_v3_41 = [
        IDSPath("source/ggd/ion/momentum"),
        IDSPath("source/ggd/ion/state/momentum"),
        IDSPath("source/ggd/neutral/momentum"),
        IDSPath("source/ggd/neutral/state/momentum"),
        IDSPath("source/ggd/momentum"),
    ]

    assert set(scalar_array_paths) == set(es_scalar_v3_41)
    assert set(vector_array_paths) == set(es_vector_v3_41)


def test_load_paths_from_ids_multiple_timesteps():
    """Test load_paths_from_ids when GGD contains multiple time steps"""
    ids = imas.IDSFactory(version="3.41.0").new("edge_sources")
    ps_reader = PlasmaStateReader(ids)

    ids.source.resize(1)
    ids.source[0].ggd.resize(2)
    ids.source[0].ggd[0].ion.resize(1)
    ids.source[0].ggd[1].ion.resize(1)
    ion_energy = ids.source[0].ggd[1].ion[0].energy
    ion_momentum = ids.source[0].ggd[1].ion[0].momentum
    ids.source[0].ggd[0].ion[0].state.resize(1)
    state_energy = ids.source[0].ggd[0].ion[0].state[0].energy
    state_momentum = ids.source[0].ggd[0].ion[0].state[0].momentum
    for ggd_idx in [0, 1]:
        _, _, scalar_array_paths, vector_array_paths = ps_reader.load_paths_from_ids(
            ggd_idx=ggd_idx
        )
        assert scalar_array_paths == vector_array_paths == []

    fill_scalar_quantity(ion_energy, 1, 1, 1)
    fill_vector_quantity(ion_momentum, 1, 1, 1)
    fill_scalar_quantity(state_energy, 1, 1, 1)
    fill_vector_quantity(state_momentum, 1, 1, 1)
    _, _, scalar_array_paths, vector_array_paths = ps_reader.load_paths_from_ids(
        ggd_idx=0
    )
    assert scalar_array_paths == [state_energy.metadata.path]
    assert vector_array_paths == [state_momentum.metadata.path]

    _, _, scalar_array_paths, vector_array_paths = ps_reader.load_paths_from_ids(
        ggd_idx=1
    )
    assert scalar_array_paths == [ion_energy.metadata.path]
    assert vector_array_paths == [ion_momentum.metadata.path]



from vtkmodules.vtkCommonCore import vtkPoints

from imas_paraview.io.read_geom import (
    convert_grid_subset_geometry_to_unstructured_grid,
    fill_vtk_points,
)
from imas_paraview.util import get_grid_ggd


def test_fill_vtk_points(ids_name, dummy_ids):
    """Tests filling VTK points from the IDS grid."""

    space_idx = 0
    points = vtkPoints()
    grid_ggd = get_grid_ggd(dummy_ids)
    fill_vtk_points(grid_ggd, space_idx, points, ids_name)
    assert points.GetNumberOfPoints() > 0


def test_convert_grid_subset_geometry_to_unstructured_grid(ids_name, dummy_ids):
    """Tests grid subset geometry conversion to unstructured grid."""

    subset_idx = 0
    space_idx = 0
    grid_ggd = get_grid_ggd(dummy_ids)
    points = vtkPoints()
    fill_vtk_points(grid_ggd, space_idx, points, ids_name)
    grid = convert_grid_subset_geometry_to_unstructured_grid(
        grid_ggd, subset_idx, points
    )
    assert grid.GetPoints() is not None



import logging
from typing import Optional

import numpy as np
from vtkmodules.vtkCommonCore import vtkPoints
from vtkmodules.vtkCommonDataModel import vtkCellArray, vtkLine, vtkPolyData

logger = logging.getLogger("imas_paraview")


def format_units(node) -> str:
    """Return the unit of the node surrounded by square brackets."""
    return f"[{node.metadata.units}]"


def iter_metadata_tree(meta):
    for child in meta._children.values():
        yield child
        yield from iter_metadata_tree(child)


def get_ggd_grid_path(ids_metadata) -> Optional[str]:
    """Finds the path of the GGD grid node within IDS metadata.

    Args:
        ids_metadata: The metadata of an IDS.

    Returns:
        The path string of the GGD grid node, or None if not found.
    """
    # Find the DD node defining the GGD grid:
    for node in iter_metadata_tree(ids_metadata):
        structure_reference = getattr(node, "structure_reference", None)
        if structure_reference in ["generic_grid_dynamic", "generic_grid_aos3_root"]:
            if getattr(node, "lifecycle_status", None) != "obsolescent":
                return node.path_string
    return None  # There are no GGD grids inside this IDS


def get_ggd_path(ids_metadata) -> Optional[str]:
    """Finds the path of the GGD node within IDS metadata.

    Args:
        ids_metadata: The metadata of an IDS.

    Returns:
        The path string of the GGD node, or None if not found.
    """
    for node in iter_metadata_tree(ids_metadata):
        metadata_name = getattr(node, "name", None)

        # Check for ggd in name metadata
        if metadata_name == "ggd":
            return node.path_string
    return None


def get_grid_ggd(ids, ggd_idx=0, parent_idx=0):
    """Finds and returns the grid_ggd within IDS at the time index ggd_idx. If the
    grid_ggd at ggd_idx time index does not exist, it tries to return the first
    grid_ggd. If this does not exist, it returns None.

    Args:
        ids: The IDS for which to return the grid_gdd.
        ggd_idx: Time index for which to load the grid.
        parent_idx: Index for any non-time-dependent parent Array of Structures.
            For example ``description_ggd[parent_idx].grid_ggd[ggd_idx]`` in the wall
            IDS.

    Returns:
        The grid_ggd node found, or None if not found.
    """
    grid_path = get_ggd_grid_path(ids.metadata)
    if grid_path is None:
        return None

    node = ids
    for path in grid_path.split("/"):
        node = node[path]
        if node.metadata.ndim == 0:
            pass  # Current node is a structure
        elif node.metadata.coordinate1.is_time_coordinate:
            # Time dependent array of structure
            if 0 <= ggd_idx < len(node):
                node = node[ggd_idx]
            else:
                node = node[0]
                logger.warning(
                    f"The GGD grid was not found at time index {ggd_idx}, so first "
                    "grid was loaded instead."
                )
        else:
            node = node[parent_idx]

    return node


def create_first_grid(ids):
    """Creates and returns the first grid_ggd within IDS.

    Args:
        ids: The IDS for which to create and return the first grid_gdd.

    Returns:
        The created first grid_gdd, or None if grid_path is None.
    """
    grid_path = get_ggd_grid_path(ids.metadata)
    if grid_path is None:
        return None

    node = ids
    for path in grid_path.split("/"):
        node = node[path]
        try:
            if len(node) == 0:
                node.resize(1)
                node = node[0]
        except TypeError:
            pass  # This was a structure and not an AoS

    return node


def create_first_ggd(ids):
    """Creates and returns the first GGD within IDS.

    Args:
        ids: The IDS for which to create and return the first GGD.

    Returns:
        The created first GGD, or None if ggd_path is None.
    """
    ggd_path = get_ggd_path(ids.metadata)
    if ggd_path is None:
        return None

    node = ids
    for path in ggd_path.split("/"):
        node = node[path]
        try:
            if len(node) == 0:
                node.resize(1)
                node = node[0]
            # Node was already resized
            elif len(node) == 1:
                node = node[0]
        except TypeError:
            pass  # This was a structure and not an AoS

    return node


def int32array(int_list):
    """Converts a list of ints to a numpy array containing values of type int32.

    Args:
        int_list: List of int64s to be converted

    Returns:
        List of converted int32s
    """
    return np.array(int_list, dtype=np.int32)


def find_closest_indices(values_to_extract, source_array):
    """Find indices of the closest values in source_array that are less than or equal
    to each value in values_to_extract.

    Args:
        values_to_extract: Values to find in the source array.
        source_array: Array to search for closest values.

    Returns:
        Indices of closest values in source_array for each value in values_to_extract.
    """
    indices = np.searchsorted(source_array, values_to_extract, side="right") - 1
    return list(indices[indices >= 0])


def pol_to_cart(rho, phi):
    """Convert from polar (or cylindrical) coordinates to cartesian.

    Args:
        rho: the radial distance
        phi: the azimuth angle in radians

    Returns:
        Tuple containing the x and y coordinates
    """
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return (x, y)


def points_to_vtkpoly(points, is_closed=False):
    """Convert a list of 3D points to a vtkPoints and vtkCellArray, which are combined
    into a single VtkPolyData object. The expected format of the points is:
    [(x1,y1,z1),(x2,y2,z2),...].

    Args:
        points: A list of 3D points, containing the X,Y,Z-coordinates in tuples of the
            form (x,y,z).
        is_closed: Boolean flag whether to close the contour. If set to true, the first
            and last point are connected by a vtkLine.

    Returns:
        vtkPolyData containing the vtkPoints and vtkLines.
    """
    vtk_points = vtkPoints()
    vtk_lines = vtkCellArray()
    for i, point in enumerate(points):
        vtk_points.InsertNextPoint(*point)
        line = vtkLine()

        if i != len(points) - 1:
            line.GetPointIds().SetId(0, i)
            line.GetPointIds().SetId(1, i + 1)
            vtk_lines.InsertNextCell(line)

    # Close loop if the points are closed
    if is_closed:
        line = vtkLine()
        line.GetPointIds().SetId(0, len(points) - 1)
        line.GetPointIds().SetId(1, 0)
        vtk_lines.InsertNextCell(line)

    vtk_poly = vtkPolyData()
    vtk_poly.SetPoints(vtk_points)
    vtk_poly.SetLines(vtk_lines)
    return vtk_poly



"""
A collection of methods to read a grid_ggd IDS node into a VTK dataset.
These methods copy contents from the grid_ggd/space and grid_ggd/grid_subset
children into distinct vtkUnstructuredGrid objects.
"""

import logging
from typing import Any, Callable

from vtkmodules.vtkCommonCore import vtkIdList, vtkPoints
from vtkmodules.vtkCommonDataModel import (
    VTK_EMPTY_CELL,
    VTK_LINE,
    VTK_POLY_LINE,
    VTK_POLYGON,
    VTK_POLYHEDRON,
    VTK_QUAD,
    VTK_TRIANGLE,
    VTK_VERTEX,
    vtkUnstructuredGrid,
)

logger = logging.getLogger("imas_paraview")


def convert_grid_subset_geometry_to_unstructured_grid(
    grid_ggd, subset_idx: int, vtk_grid_points, progress=None
) -> vtkUnstructuredGrid:
    """Copy the elements found in given grid_ggd/grid_subset IDS node into a
    vtkUnstructuredGrid instance. This method uses the supplied point coordinates in
    the form of a vtkPoints instance.
    Args:
        grid_ggd: a grid_ggd ids node
        subset_idx: an index into grid_ggd/grid_subset
        vtk_grid_points: the point coordinates corresponding to 1d objects in
        the subset elements.
        progress: Progress indicator for Paraview.


    Returns:
        The vtkUnstructuredGrid containing the given points.
    """
    grid = vtkUnstructuredGrid()
    grid.SetPoints(vtk_grid_points)
    if subset_idx >= 0:
        _fill_vtk_cell_array_from_gs(grid_ggd, subset_idx, grid, progress)
    else:
        _fill_vtk_cell_array_from_gs2(grid_ggd, subset_idx, grid, progress)
    return grid


def fill_vtk_points(
    grid_ggd, space_idx: int, points: vtkPoints, ids_name: str, progress=None
) -> None:
    """Populate the vtkPoints data structure with coordinates from the grid_ggd/space
    IDS node.

    Args:
        grid_ggd: a grid_ggd ids node.
        space_idx: an index into the grid_ggd/space AoS.
        points: the vtk points instance.
        ids_name: The name of the IDS.
        progress: Progress indicator for Paraview.
    """
    num_objects0d = len(grid_ggd.space[space_idx].objects_per_dimension[0].object)
    logger.info(f"Reading {num_objects0d} points from grid_ggd/space[{space_idx}]")

    s = 1  # scale objects from mm to m
    if len(grid_ggd.space[space_idx].objects_per_dimension[0].object[0].geometry) == 0:
        raise RuntimeError("Geometry of object is empty.")
    if grid_ggd.space[space_idx].objects_per_dimension[0].object[0].geometry[0] > 100:
        s = 0.001

    points.Allocate(num_objects0d, 0)
    coordinate_type = grid_ggd.space[space_idx].coordinates_type

    # When length of coordinate type is greater than 2, the third coordinate is in
    # geometry[2], else it is 0.
    if len(coordinate_type) > 2:
        third_dim: Callable[[Any], int] = lambda e: getattr(e, "geometry")[2]  # noqa
    else:
        third_dim: Callable[[Any], int] = lambda e: 0  # noqa

    objects = grid_ggd.space[space_idx].objects_per_dimension[0].object
    for obj in objects:
        if progress:
            progress.increment(0.5 / len(objects))
        if ids_name == "wall":
            points.InsertNextPoint(
                (obj.geometry[0] * s, obj.geometry[1] * s, third_dim(obj) * s)
            )
        else:
            if len(obj.geometry) < 2:
                raise RuntimeError("Geometry of object is smaller than 2.")
            points.InsertNextPoint(
                (obj.geometry[0] * s, third_dim(obj) * s, obj.geometry[1] * s)
            )
            # Use for changing orientation in paraview.


def _fill_vtk_cell_array_from_gs2(
    grid_ggd, subset_idx: int, ugrid: vtkUnstructuredGrid, progress=None
) -> None:
    """_fill_vtk_cell_array_from_gs() for wall IDS.

    Args:
        grid_ggd: a grid_ggd ids node.
        subset_idx: -1
        ugrid: the vtk unstructured grid instance.
        progress: Progress indicator for Paraview.
    """
    grid = grid_ggd.space[0].objects_per_dimension
    num_cell = len(grid[2].object)  # 4- 0D, 1D, 2D, 3D objects
    ugrid.AllocateEstimate(num_cell, 10)

    # Uses only 2d cells.
    for j in range(len(grid[2].object)):
        if progress:
            progress.increment(0.5 / num_cell)
        obj = grid[2].object[j]
        obj_nodes = obj.nodes
        obj_dimension = 2
        # try:
        #     obj_boundary = obj.boundary
        # except:
        #     obj_boundary = []

        pt_ids = list(map(lambda val: val - 1, obj_nodes))
        npts = len(pt_ids)
        cell_type = _get_vtk_cell_type(obj_dimension, npts)
        ugrid.InsertNextCell(cell_type, npts, pt_ids)


def _fill_vtk_cell_array_from_gs(
    grid_ggd, subset_idx: int, ugrid: vtkUnstructuredGrid, progress=None
) -> None:
    """Populate the cells in the vtk unstructured grid instance with elements from the
    grid_ggd/grid_subset IDS node.

    Args:
        grid_ggd: a grid_ggd ids node.
        subset_idx: an index into the grid_ggd/grid_subset AoS.
        ugrid: the vtk unstructured grid instance.
        progress: Progress indicator for Paraview.
    """
    grid_subset = grid_ggd.grid_subset[subset_idx]
    num_gs_el = len(grid_subset.element)

    if hasattr(grid_subset, "identifier"):
        logger.info(f"Reading {num_gs_el} elements from {grid_subset.identifier.name}")
    else:
        logger.info(
            f"Reading {num_gs_el} elements from grid_ggd/grid_subset[{subset_idx}]"
        )

    ugrid.AllocateEstimate(num_gs_el, 10)
    object_3d_pt_ids = vtkIdList()

    for element in grid_subset.element:
        if progress:
            progress.increment(0.5 / num_gs_el)
        for object_ in element.object:
            obj_space = object_.space - 1
            obj_index = object_.index - 1
            obj_dimension = object_.dimension - 1
            obj_nodes = (
                grid_ggd.space[obj_space]
                .objects_per_dimension[obj_dimension]
                .object[obj_index]
                .nodes
            )
            obj_boundary = (
                grid_ggd.space[obj_space]
                .objects_per_dimension[obj_dimension]
                .object[obj_index]
                .boundary
            )

            pt_ids = list(
                map(lambda val: val - 1, obj_nodes)
            )  # offset by -1 as fortran indexing used in IMAS( 1,...n)
            npts = len(pt_ids)
            cell_type = _get_vtk_cell_type(obj_dimension, npts)

            if cell_type != VTK_POLYHEDRON:
                ugrid.InsertNextCell(cell_type, npts, pt_ids)

            elif cell_type == VTK_POLYHEDRON:
                num_faces = len(obj_boundary)
                object_3d_pt_ids.Reset()
                object_3d_pt_ids.InsertNextId(num_faces)

                for f in range(num_faces):
                    object_2d_idx = obj_boundary[f].index - 1
                    object_2d = (
                        grid_ggd.space[obj_space]
                        .objects_per_dimension[2]
                        .object[object_2d_idx]
                    )
                    object_2d_pt_ids = list(
                        map(lambda val: val - 1, object_2d.nodes)
                    )  # offset by -1
                    num_face_pts = len(object_2d_pt_ids)

                    # the format for 3d cell point ids is
                    #   [numFace0Pts, p0, p1, .., numFace1Pts, p0, p1, ...]
                    object_3d_pt_ids.InsertNextId(num_face_pts)

                    for pt in object_2d_pt_ids:
                        object_3d_pt_ids.InsertNextId(pt)

                # the other overload (int, int, (int,...)) raises TypeError,
                # so use the one with vtkIdList
                ugrid.InsertNextCell(cell_type, object_3d_pt_ids)


def _get_vtk_cell_type(dimension: int, npts: int) -> int:
    """Determines a suitable VTK cell type from given cell dimensionality and number of
    points for that cell.

    Args:
        dimension: the number of dimensions for a cell.
        npts: the number of points for a cell.

    Returns:
        VTK Cell Type
    """
    if dimension == 0:
        return VTK_VERTEX

    elif dimension == 1:
        if npts == 2:
            return VTK_LINE
        else:
            return VTK_POLY_LINE

    elif dimension == 2:
        if npts == 3:
            return VTK_TRIANGLE
        elif npts == 4:
            return VTK_QUAD
        else:
            return VTK_POLYGON

    elif dimension == 3:
        return VTK_POLYHEDRON

    else:
        return VTK_EMPTY_CELL



import imas
import numpy as np
import pytest
from imas import DBEntry
from imas.backends.imas_core.db_entry_helpers import IDS_TIME_MODE_HOMOGENEOUS
from vtk.util.numpy_support import vtk_to_numpy
from vtkmodules.vtkCommonDataModel import vtkPartitionedDataSetCollection

from imas_paraview.plugins.profiles_2d import Profiles2DReader


@pytest.mark.skip(reason="no IMAS-Core available")
def test_load_profiles():
    """Test if 2D profiles structures are loaded in the VTK
    PartitionedDatasetCollection."""
    reader = Profiles2DReader()

    with DBEntry(
        "imas:hdf5?path=/work/imas/shared/imasdb/ITER_SCENARIOS/3/110004/1",
        "r",
    ) as entry:
        ids = entry.get("equilibrium", lazy=True, autoconvert=False)
        reader._ids = ids
        reader.setup_ids()

        profile1 = ids.time_slice[0].profiles_2d[0].psi
        name1 = "Psi"
        profile2 = ids.time_slice[0].profiles_2d[0].j_tor
        name2 = "J_tor"

        # 1 selection
        output = vtkPartitionedDataSetCollection()
        reader._selected = [name1]
        reader._load_profiles(output)

        assert output.GetNumberOfPartitionedDataSets() == 1
        check_ugrid(output.GetPartitionedDataSet(0).GetPartition(0), profile1, reader)

        # 2 selections
        output = vtkPartitionedDataSetCollection()
        reader._selected = [name1, name2]
        reader._load_profiles(output)

        assert output.GetNumberOfPartitionedDataSets() == 2
        check_ugrid(output.GetPartitionedDataSet(0).GetPartition(0), profile1, reader)
        check_ugrid(output.GetPartitionedDataSet(1).GetPartition(0), profile2, reader)


def check_ugrid(ugrid, expected_profile, reader):
    """assert that the data in the ugrid matches the IDS profiles arrays."""
    num_points = ugrid.GetNumberOfPoints()
    vtk_scalars = ugrid.GetPointData().GetScalars()
    scalars_array = vtk_to_numpy(vtk_scalars)
    points_np = np.array([ugrid.GetPoint(i) for i in range(num_points)])
    r_vtk, z_vtk = points_np[:, 0], points_np[:, 2]

    assert np.array_equal(expected_profile.ravel(), scalars_array)
    assert np.array_equal(reader.r.ravel(), r_vtk)
    assert np.array_equal(reader.z.ravel(), z_vtk)


def test_load_profiles_dummy_equilibrium():
    ids = imas.IDSFactory(version="4.0.0").new("equilibrium")
    ids.time = [0]
    ids.ids_properties.homogeneous_time = IDS_TIME_MODE_HOMOGENEOUS
    ids.time_slice.resize(1)
    ids.time_slice[0].profiles_2d.resize(1)
    profiles_2d = ids.time_slice[0].profiles_2d[0]
    profiles_2d.r = np.array([[1.1, 2.2], [3.3, 4.4]])
    profiles_2d.z = np.array([[5.5, 6.6], [7.7, 8.8]])
    profiles_2d.psi = np.array([[10.1, 10.2], [10.3, 10.4]])

    reader = Profiles2DReader()
    reader._ids = ids
    reader.setup_ids()
    output = vtkPartitionedDataSetCollection()
    reader._selected = ["Psi"]
    reader._load_profiles(output)
    assert output.GetNumberOfPartitionedDataSets() == 1
    profile = ids.time_slice[0].profiles_2d[0].psi
    check_ugrid(output.GetPartitionedDataSet(0).GetPartition(0), profile, reader)


def test_load_profiles_dummy_core_profiles():
    ids = imas.IDSFactory(version="4.0.0").new("core_profiles")
    ids.time = [0]
    ids.ids_properties.homogeneous_time = IDS_TIME_MODE_HOMOGENEOUS
    ids.profiles_2d.resize(1)
    ids.profiles_2d[0].ion.resize(1)
    profiles_2d = ids.profiles_2d[0]
    profiles_2d.grid.dim1 = [1, 2]
    profiles_2d.grid.dim2 = [3, 4]
    profiles_2d.grid_type = 1
    profiles_2d.ion[0].pressure = np.array([[10.1, 10.2], [10.3, 10.4]])

    reader = Profiles2DReader()
    reader._ids = ids
    reader.setup_ids()
    output = vtkPartitionedDataSetCollection()
    reader._selected = ["Ion Pressure"]
    reader._load_profiles(output)
    assert output.GetNumberOfPartitionedDataSets() == 1
    profile = profiles_2d.ion[0].pressure
    check_ugrid(output.GetPartitionedDataSet(0).GetPartition(0), profile, reader)



"""Plugin to view beam structures of ec_launchers IDS in Paraview."""

import logging
from dataclasses import dataclass

import numpy as np
from imas.ids_structure import IDSStructure
from paraview.util.vtkAlgorithm import smhint, smproxy
from vtkmodules.vtkCommonDataModel import vtkMultiBlockDataSet

from imas_paraview.ids_util import get_object_by_name
from imas_paraview.paraview_support.servermanager_tools import (
    doublevector,
    propertygroup,
)
from imas_paraview.plugins.base_class import GGDVTKPluginBase
from imas_paraview.util import find_closest_indices, points_to_vtkpoly, pol_to_cart

logger = logging.getLogger("imas_paraview")


@dataclass
class Beam:
    """Data class that stores ``beam`` IDS structures, along with its name."""

    name: str
    beam: IDSStructure


@smproxy.source(label="Beam Reader")
@smhint.xml("""<ShowInMenu category="IMAS Tools" />""")
class BeamReader(GGDVTKPluginBase, is_time_dependent=True):
    def __init__(self):
        super().__init__("vtkMultiBlockDataSet", ["ec_launchers"])
        self.distance = 10

    @doublevector(label="Beam Distance", name="beam_distance", default_values=10)
    def P99_SetDistance(self, val):
        """Sets the beam's distance from the launching position."""
        self._update_property("distance", val)

    @propertygroup("Beam settings", ["beam_distance"])
    def PG3_BeamGroup(self):
        """Dummy function to define a PropertyGroup."""

    def GetAttributeArrayName(self, idx) -> str:
        return self._selectable[idx].name

    def RequestData(self, request, inInfo, outInfo):
        if self._dbentry is None or not self._ids_and_occurrence or self._ids is None:
            return 1

        # Retrieve the selected time step and profiles
        time = self._get_selected_time_step(outInfo)
        if time is None:
            logger.warning("Selected invalid time step")
            return 1

        index_list = find_closest_indices([time], self._ids.time)
        time_idx = index_list[0]

        if len(self._selected) > 0:
            output = vtkMultiBlockDataSet.GetData(outInfo)
            self._load_beam(output, time_idx)
        return 1

    def setup_ids(self):
        """
        Select which beams to show in the array domain selector. The IDS is searched
        for beam structures and their names are added to the array domain selector.
        """
        assert self._ids is not None, "IDS cannot be empty during setup."

        self._selectable = []

        beams = self._ids.beam

        for i, beam in enumerate(beams):
            beam_name = beam.name
            if beam_name == "":
                beam_name = f"beam {i}"
                logger.warning(
                    f"Found a channel without a name, it will be loaded as {beam_name}."
                )

            selectable = Beam(str(beam_name), beam)
            self._selectable.append(selectable)

    def _load_beam(self, output, time_idx):
        """Go through the list of selected beams, and load each of them in a
        separate vtkPolyData object, which are all combined into a vtkMultiBlockDataSet.

        Args:
            output: The vtkMultiBlockDataSet containing the beams.
        """
        for i, beam_name in enumerate(self._selected):
            beam = get_object_by_name(self._selectable, beam_name)
            if beam is None:
                raise ValueError(f"Could not find {beam_name}")

            logger.info(f"Selected {beam.name}")
            vtk_poly = self._create_vtk_beam(beam.beam, time_idx)
            output.SetBlock(i, vtk_poly)

    def _create_vtk_beam(self, beam, time_idx):
        """Create a vtkPolyData containing two points from the beam's
        launching_position to a point in the direction that the steering angles point
        to, at a distance from the launching position. For this, the points and
        direction are converted to cartesian coordinates, and a vtkLine is created
        between these two points.

        Args:
            beam: the beam structure to create.

        Returns:
            vtkPolyData containing the beams
        """

        first_point, second_point = self.transform_points(
            beam.launching_position.r[time_idx],
            beam.launching_position.phi[time_idx],
            beam.launching_position.z[time_idx],
            beam.steering_angle_pol[time_idx],
            beam.steering_angle_tor[time_idx],
        )

        points = [first_point, second_point]
        vtk_poly = points_to_vtkpoly(points)
        return vtk_poly

    def transform_points(
        self,
        launch_pos_r,
        launch_pos_phi,
        launch_pos_z,
        steering_angle_pol,
        steering_angle_tor,
    ):
        """Transform the launching position and launching direction to cartesian
        coordinates.

        Args:
            launch_pos_r: Radial distance of the launching position
            launch_pos_phi: Azimuth angle of the launching position
            launch_pos_z: Height of the launching position
            steering_angle_pol: Steering angle of the beam in the R,Z plane
            steering_angle_tor: Steering angle of the beam away from the poloidal plane
            distance: Distance along direction vector at which to place the second point

        Returns:
            tuple containing the launching position and a point into the direction of
            the launch, in cartesian coordinates: ((x1,y1,z1),(x2,y2,z2))
        """

        first_point = (*pol_to_cart(launch_pos_r, launch_pos_phi), launch_pos_z)

        # Calculate radial components of the wave vector (taking |k|=1)
        k_r = -np.cos(steering_angle_pol) * np.cos(steering_angle_tor)
        k_phi = np.cos(steering_angle_pol) * np.sin(steering_angle_tor)

        # Convert to direction in cartesian coordinates
        k_x = k_r * np.cos(launch_pos_phi) - k_phi * np.sin(launch_pos_phi)
        k_y = k_r * np.sin(launch_pos_phi) + k_phi * np.cos(launch_pos_phi)
        k_z = -np.sin(steering_angle_pol)

        # Calculate the end point from the launching position into the direction of the
        # steering angles
        second_point = (
            first_point[0] + self.distance * k_x,
            first_point[1] + self.distance * k_y,
            first_point[2] + self.distance * k_z,
        )

        return first_point, second_point



import logging

import numpy as np
from paraview.util.vtkAlgorithm import smdomain, smhint, smproperty, smproxy
from vtk.util.numpy_support import vtk_to_numpy
from vtkmodules.numpy_interface import dataset_adapter as dsa
from vtkmodules.util.vtkAlgorithm import VTKPythonAlgorithmBase
from vtkmodules.vtkCommonDataModel import vtkPartitionedDataSetCollection, vtkTable

from imas_paraview.paraview_support.servermanager_tools import (
    arrayselectiondomain,
    arrayselectionstringvector,
)
from imas_paraview.progress import Progress

logger = logging.getLogger("imas_paraview")


@smproxy.filter(label="1D Profiles Mapper")
@smhint.xml("""<ShowInMenu category="IMAS Tools" />""")
@smproperty.input(name="1D Profile", port_index=1)
@smdomain.datatype(dataTypes=["vtkTable"], composite_data_supported=False)
@smproperty.input(name="Psi Grid", port_index=0)
@smdomain.datatype(
    dataTypes=["vtkPartitionedDataSetCollection"], composite_data_supported=True
)
class Profiles1DMapper(VTKPythonAlgorithmBase):
    """Filter that maps 1d profiles onto a 2D poloidal flux grid. This filter takes
    2 inputs:
    - A poloidal flux (Psi) from a GGD. i.e. from GGD Reader/JOREK Reader (input 0)
    - A 1D profile from the 1DProfilesReader. (input 1)
    The 1DProfilesReader must output at least the Psi grid coordinate, and can output
    any number of 1D profiles."""

    def __init__(self):
        VTKPythonAlgorithmBase.__init__(
            self,
            nInputPorts=2,
            nOutputPorts=1,
            outputType="vtkPartitionedDataSetCollection",
        )
        self._selected = []
        self._selectable = []

    def FillInputPortInformation(self, port, info):
        if port == 1:
            info.Set(self.INPUT_REQUIRED_DATA_TYPE(), "vtkTable")
        else:
            info.Set(self.INPUT_REQUIRED_DATA_TYPE(), "vtkPartitionedDataSetCollection")
        return 1

    def RequestData(self, request, inInfoVec, outInfo):
        input0 = dsa.WrapDataObject(
            vtkPartitionedDataSetCollection.GetData(inInfoVec[0], 0)
        )
        input1 = dsa.WrapDataObject(vtkTable.GetData(inInfoVec[1], 0))

        self._selectable = input1.RowData.keys()
        psi = input0.PointData["Psi [Wb]"]

        if isinstance(psi, dsa.VTKNoneArray):
            logger.warning(
                "The GGD Reader should output a poloidal flux GGD. Please select 'Psi' "
                "in the attribute selector window."
            )
            return 1
        psi_grid = psi.GetArrays()[0]
        psi_profiles = input1.RowData["Grid Psi"]

        if isinstance(psi_profiles, dsa.VTKNoneArray) and self._selected:
            logger.warning(
                "The 1DProfilesReader should output a poloidal flux grid. Please"
                " select 'Grid Psi' in the attribute selector window of the "
                "1DProfilesReader."
            )
            return 1

        output = dsa.WrapDataObject(vtkPartitionedDataSetCollection.GetData(outInfo))
        output.ShallowCopy(input0.VTKObject)

        progress = Progress(self.UpdateProgress)

        for profile_name in self._selected:
            profile = input1.RowData[profile_name]

            xp = vtk_to_numpy(psi_profiles)
            fp = vtk_to_numpy(profile)

            # numpy's interp requires xp to be strictly increasing, and the DD does not
            # enforce a specific ordering. So we invert the array if it is decreasing
            if xp[0] > xp[-1]:
                xp = xp[::-1]
                fp = fp[::-1]

            psi_grid_values = vtk_to_numpy(psi_grid)
            resample = np.interp(psi_grid_values, xp, fp, left=np.nan, right=np.nan)

            progress.increment(1 / len(self._selectable))
            output.PointData.append(resample, f"{profile_name} (resampled)")
        return 1

    @arrayselectiondomain(
        property_name="Profile1DArray",
        name="Profile1DSelector",
        label="Select 1D Profiles",
    )
    def SetProfile1D(self, array, status):
        """Select all or a subset of available 1d_profiles to load."""
        # Add an array to selected list
        if status == 1 and array not in self._selected:
            self._selected.append(array)
            self.Modified()

        # Remove an array from selected list
        if status == 0 and array in self._selected:
            self._selected.remove(array)
            self.Modified()

    @arrayselectionstringvector(
        property_name="Profile1DArray", attribute_name="Profile1D"
    )
    def _Profile1DArraySelector(self):
        pass

    def GetNumberOfProfile1DArrays(self):
        return len(self._selectable)

    def GetProfile1DArrayName(self, idx) -> str:
        return self._selectable[idx]

    def GetProfile1DArrayStatus(self, *args):
        return 1
